from os import path
from pathlib import Path
from typing import List, Literal

import astropy.units as u
import numpy as np
import pytest
from astropy import wcs
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time

from ..cube_cut import CutoutFactory
from ..exceptions import InputWarning, InvalidQueryError
from ..make_cube import CubeFactory, TicaCubeFactory
from .utils_for_test import create_test_ffis


@pytest.fixture
def ffi_files(ffi_type: Literal["SPOC", "TICA"], tmp_path):
    """Pytest fixture for creating test ffi files"""

    tmpdir = str(tmp_path)

    img_sz = 10
    num_im = 100

    return create_test_ffis(img_sz, num_im, dir_name=tmpdir, product=ffi_type)


@pytest.fixture
def cube_file(ffi_type: Literal["SPOC", "TICA"], ffi_files: List[str], tmp_path):
    """Pytest fixture for creating a cube file"""

    tmpdir = str(tmp_path)

    # Making the test cube
    if ffi_type == "SPOC":
        cube_maker = CubeFactory()
    else:
        cube_maker = TicaCubeFactory()

    cube_file = cube_maker.make_cube(ffi_files, path.join(tmpdir, "test_cube.fits"), verbose=False)

    return cube_file


def checkcutout(product, cutfile, pixcrd, world, csize, ecube, eps=1.e-7):
    """Check FITS cutout for correctness
    
    Checks RA_OBJ/DEC_OBJ in primary header, and TIME, FLUX, and
    FLUX_ERR in table.
    
    Inputs:
    cutfile  Name of FITS cutout file
    pixcrd   [2] pixel coordinates for cutout center [cy,cx]
    world    [2] RA, Dec in degrees for cutout center
    csize    Integer size of cutout (probably should be odd)
    ecube    Simulated data cube
    eps      Maximum allowed distance offset in degrees
    Returns True on success, False on failure
    """
    
    ix = int(pixcrd[1])
    iy = int(pixcrd[0])
    x1 = ix - csize//2
    x2 = x1 + csize - 1
    y1 = iy - csize//2
    y2 = y1 + csize - 1
    hdulist = fits.open(cutfile)
    ra_obj = hdulist[0].header['RA_OBJ']
    dec_obj = hdulist[0].header['DEC_OBJ']
    pinput = SkyCoord(world[0], world[1], frame='icrs', unit='deg')
    poutput = SkyCoord(ra_obj, dec_obj, frame='icrs', unit='deg')
    
    dist = pinput.separation(poutput).degree
    assert dist <= eps, "{} separation in primary header {} too large".format(cutfile, dist)
        
    ntimes = ecube.shape[2]
    tab = hdulist[1].data
    assert len(tab) == ntimes, "{} expected {} entries, found {}".format(cutfile, ntimes, len(tab))
    assert (tab['TIME'] == (np.arange(ntimes)+0.5)).all(), "{} some time values are incorrect".format(cutfile)

    if product == 'SPOC':
        # TODO: Modify check1 to take TICA - adjust for TICA's slightly different WCS
        # solutions (TICA will usually be ~1 pixel off from SPOC for the same cutout)
        check1(tab['FLUX'], x1, x2, y1, y2, ecube[:, :, :, 0], 'FLUX', cutfile)
        # Only SPOC propagates errors, so TICA will always have an empty error array
        check1(tab['FLUX_ERR'], x1, x2, y1, y2, ecube[:, :, :, 1], 'FLUX_ERR', cutfile)
    
    # Regression test for PR #6
    assert hdulist[2].data.dtype.type == np.int32

    return 


def check1(flux, x1, x2, y1, y2, ecube, label, cutfile):
    """ Checking to make sure the right corresponding pixels 
    are replaced by NaNs when cutout goes off the TESS camera.
    Test one of flux or error
    """

    cx = ecube.shape[0]
    cy = ecube.shape[1]

    if x1 < 0:
        assert np.isnan(flux[:, :-x1, :]).all(), "{} {} x1 NaN failure".format(cutfile, label)
    
    if y1 < 0:
        assert np.isnan(flux[:, :, :-y1]).all(), "{} {} y1 NaN failure".format(cutfile, label)
        
    if x2 >= cx:
        assert np.isnan(flux[:, -(x2-cx+1):, :]).all(), "{} {} x2 NaN failure".format(cutfile, label)
        
    if y2 >= cy:
        assert np.isnan(flux[:, :, -(y2-cy+1):]).all(), "{} {} y2 NaN failure".format(cutfile, label)
        
    x1c = max(x1, 0)
    y1c = max(y1, 0)
    x2c = min(x2, cx-1)
    y2c = min(y2, cy-1)

    scube = ecube[x1c:x2c, y1c:y2c, :]
    sflux = np.moveaxis(flux[:, x1c-x1:x2c-x1, y1c-y1:y2c-y1], 0, -1)

    assert (scube == sflux).all(), "{} {} comparison failure".format(cutfile, label)


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_cube_cutout(cube_file, ffi_files, ffi_type, tmp_path):
    """
    Testing the cube cutout functionality.
    """

    tmpdir = str(tmp_path)

    img_sz = 10
    num_im = 100

    cutout_maker = CutoutFactory()

    # Read one of the input images to get the WCS
    n = 1 if ffi_type == 'SPOC' else 0
    img_header = fits.getheader(ffi_files[0], n)
    cube_wcs = wcs.WCS(img_header)

    # get pixel positions at edges and center of image
    # somewhat cryptic one-liner to get the grid of points
    pval = np.array([0, img_sz // 2, img_sz - 1], dtype=float)
    pixcrd = pval[np.transpose(np.reshape(np.mgrid[0:3, 0:3], (2, 9)))]
    
    # add one more giant cutout that goes off all 4 edges
    pixcrd = np.append(pixcrd, pixcrd[4].reshape(1, 2), axis=0)

    # getting the world coordinates
    world_coords = cube_wcs.all_pix2world(pixcrd, 0)

    # Getting the cutouts
    cutbasename = 'make_cube_cutout_{}.fits'
    cutlist = [path.join(tmpdir, cutbasename.format(i)) for i in range(len(world_coords))]
    csize = [img_sz//2]*len(world_coords)
    csize[-1] = img_sz+5
    for i, v in enumerate(world_coords):
        coord = SkyCoord(v[0], v[1], frame='icrs', unit='deg')
        cutout_maker.cube_cut(cube_file, coord, csize[i], ffi_type, target_pixel_file=cutlist[i],
                              output_path=tmpdir, verbose=False)

    # expected values for cube
    ecube = np.zeros((img_sz, img_sz, num_im, 2))
    plane = np.arange(img_sz*img_sz, dtype=np.float32).reshape((img_sz, img_sz))
    for i in range(num_im):
        ecube[:, :, i, 0] = -plane
        ecube[:, :, i, 1] = plane
        plane += img_sz*img_sz

    # Doing the actual checking
    for i, cutfile in enumerate(cutlist):
        checkcutout(ffi_type, cutfile, pixcrd[i], world_coords[i], csize[i], ecube)


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_parse_table_info(cube_file, ffi_type, tmp_path):
    """Test _parse_table_info"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    coord = "256.88 6.38"
    cutout_size = [5, 3]
    out_file = cutout_maker.cube_cut(
        cube_file, coord, cutout_size, ffi_type, output_path=path.join(tmpdir, "out_dir"), verbose=False
    )

    assert "256.880000_6.380000_5x3_astrocut.fits" in out_file

    assert isinstance(cutout_maker.cube_wcs, wcs.WCS)
    ra, dec = cutout_maker.cube_wcs.wcs.crval
    assert round(ra, 4) == 250.3497
    assert round(dec, 4) == 2.2809

    # checking on the center coordinate too
    coord = SkyCoord(256.88, 6.38, frame="icrs", unit="deg")
    assert cutout_maker.center_coord.separation(coord) == 0


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_header_keywords_quality(cube_file, ffi_type, tmp_path):
    """Test header keywords quality"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    coord = "256.88 6.38"
    cutout_size = [5, 3]
    out_file = cutout_maker.cube_cut(
        cube_file, coord, cutout_size, ffi_type, output_path=path.join(tmpdir, "out_dir"), verbose=False
    )

    with fits.open(out_file) as hdulist:

        # Primary header checks
        primary_header = hdulist[0].header
        assert primary_header['FFI_TYPE'] == ffi_type
        assert primary_header['CREATOR'] == 'astrocut'
        assert primary_header['BJDREFI'] == 2457000

        if ffi_type == 'TICA':
            # Verifying DATE-OBS calculation in TICA
            date_obs = primary_header['DATE-OBS']
            tstart = Time(date_obs).jd - primary_header['BJDREFI']
            assert primary_header['TSTART'] == tstart

            # Verifying DATE-END calculation in TICA
            date_end = primary_header['DATE-END']
            tstop = Time(date_end).jd - primary_header['BJDREFI']
            assert primary_header['TSTOP'] == tstop

        # Checking for header keyword propagation in EXT 1 and 2
        ext1_header = hdulist[1].header
        ext2_header = hdulist[2].header
        assert ext1_header['FFI_TYPE'] == ext2_header['FFI_TYPE'] == primary_header['FFI_TYPE']
        assert ext1_header['CREATOR'] == ext2_header['CREATOR'] == primary_header['CREATOR']
        assert ext1_header['BJDREFI'] == ext2_header['BJDREFI'] == primary_header['BJDREFI']


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_header_keywords_diffs(cube_file, ffi_type, tmp_path):
    """Test known header keywords differences"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    coord = "256.88 6.38"
    cutout_size = [5, 3]
    out_file = cutout_maker.cube_cut(
        cube_file, coord, cutout_size, ffi_type, output_path=path.join(tmpdir, "out_dir"), verbose=False
    )

    with fits.open(out_file) as hdulist:

        # Get units from BinTableHDU
        cols = hdulist[1].columns.info('name, unit', output=False)
        cols_dict = dict(zip(*cols.values()))

        ncols = 12 if ffi_type == 'SPOC' else 11
        assert len(cols_dict) == ncols

        # Checking for known keyword differences
        if ffi_type == 'SPOC':
            assert hdulist[0].header['TIMEREF'] == 'SOLARSYSTEM', 'TIMEREF keyword does not match expected'
            assert hdulist[0].header['TASSIGN'] == 'SPACECRAFT', 'TASSIGN keyword does not match expected'
            assert cols_dict['FLUX'] == 'e-/s', f'Expected `FLUX` units of "e-/s", got units of "{cols_dict["FLUX"]}"'
            assert hdulist[1].data.field('CADENCENO').all() == 0.0

        if ffi_type == 'TICA':
            assert hdulist[0].header['TIMEREF'] is None, 'TIMEREF keyword does not match expected'
            assert hdulist[0].header['TASSIGN'] is None, 'TASSIGN keyword does not match expected'
            assert cols_dict['FLUX'] == 'e-', f'Expected `FLUX` units of "e-", got units of "{cols_dict["FLUX"]}"'
            assert hdulist[1].data.field('CADENCENO').all() != 0.0


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_get_cutout_limits(cube_file, ffi_type, tmp_path):
    """Test _get_cutout_limits"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()

    # Making the cutout
    coord = "256.88 6.38"
    cutout_size = [5, 3]
    out_file = cutout_maker.cube_cut(
        cube_file, coord, cutout_size, ffi_type, output_path=path.join(tmpdir, "out_dir"), verbose=False
    )

    xmin, xmax = cutout_maker.cutout_lims[0]
    ymin, ymax = cutout_maker.cutout_lims[1]

    assert (xmax-xmin) == cutout_size[0]
    assert (ymax-ymin) == cutout_size[1]

    cutout_size = [5*u.pixel, 7*u.pixel]
    out_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)
    assert "256.880000_6.380000_5x7_astrocut.fits" in out_file

    xmin, xmax = cutout_maker.cutout_lims[0]
    ymin, ymax = cutout_maker.cutout_lims[1]

    assert (xmax-xmin) == cutout_size[0].value
    assert (ymax-ymin) == cutout_size[1].value

    cutout_size = [3*u.arcmin, 5*u.arcmin]
    out_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)
    assert "256.880000_6.380000_8x15_astrocut.fits" in out_file

    xmin, xmax = cutout_maker.cutout_lims[0]
    ymin, ymax = cutout_maker.cutout_lims[1]

    assert (xmax - xmin) == 8
    assert (ymax - ymin) == 15


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_small_cutout(cube_file, ffi_type, tmp_path):

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    cutout_size = [1 * u.arcsec, 5 * u.arcsec]
    coord = "256.88 6.38"

    out_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)
    assert "256.880000_6.380000_1x1_astrocut.fits" in out_file

    xmin, xmax = cutout_maker.cutout_lims[0]
    ymin, ymax = cutout_maker.cutout_lims[1]

    assert (xmax - xmin) == 1
    assert (ymax - ymin) == 1


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_get_full_cutout_wcs(cube_file, ffi_type, tmp_path):
    """Test _get_full_cutout_wcs"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    cutout_size = [5, 3]
    coord = "256.88 6.38"

    cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)
    cutout_wcs_full = cutout_maker._get_full_cutout_wcs(fits.getheader(cube_file, 2))
    assert (
        cutout_wcs_full.wcs.crpix == [1045 - cutout_maker.cutout_lims[0, 0], 1001 - cutout_maker.cutout_lims[1, 0]]
    ).all()


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_fit_cutout_wcs(cube_file, ffi_type, tmp_path):
    """Test _fit_cutout_wcs"""

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    cutout_size = [5, 3]
    coord = "256.88 6.38"

    cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)
    cutout_wcs_full = cutout_maker._get_full_cutout_wcs(fits.getheader(cube_file, 2))

    max_dist, sigma = cutout_maker._fit_cutout_wcs(cutout_wcs_full, (3, 5))
    assert max_dist.deg < 1e-05
    assert sigma < 1e-05

    cry, crx = cutout_maker.cutout_wcs.wcs.crpix
    assert round(cry) == 3
    assert round(crx) == 2


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_target_pixel_file(cube_file, ffi_type, tmp_path):
    """Test target pixel file"""
    
    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()
    cutout_size = [5, 3]
    coord = "256.88 6.38"

    out_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, verbose=False, output_path=tmpdir)

    # Testing the cutout content is in test_cube_cutout
    # this tests that the format of the tpf is what it should be
    tpf = fits.open(out_file)

    assert tpf[0].header["ORIGIN"] == 'STScI/MAST'

    tpf_table = tpf[1].data
    # SPOC cutouts have 1 extra columns in EXT 1
    ncols = 12 if ffi_type == 'SPOC' else 11
    assert len(tpf_table.columns) == ncols
    assert "TIME" in tpf_table.columns.names
    assert "FLUX" in tpf_table.columns.names
    assert "FLUX_ERR" in tpf_table.columns.names
    assert "FFI_FILE" in tpf_table.columns.names

    # Check img cutout shape and data type
    cutout_img = tpf_table[0]['FLUX']
    assert cutout_img.shape == (3, 5)
    assert cutout_img.dtype.name == 'float32'

    # Check error cutout shape, contents, and data type
    cutout_err = tpf_table[0]['FLUX_ERR']
    assert cutout_err.shape == (3, 5)
    if ffi_type == "TICA":
        assert np.mean(cutout_err) == 0
    assert cutout_err.dtype.name == 'float32'

    # Check aperture shape and data type
    aperture = tpf[2].data
    assert aperture.shape == (3, 5)
    assert aperture.dtype.name == 'int32'

    tpf.close()


def test_tica_cutout_error(tmp_path):
    """
    Test cutouts created from existing TICA cubes (i.e., those with
    error arrays)

    Write test to check that cutouts can be made from *both* the TICA
    cubes that still have the error array, as well as cubes that
    *have* been updated and no longer have the error array. Needed to
    verify that CutoutFactory will work on the cubes that have not 
    been remade yet.
    """

    tmpdir = str(tmp_path)

    img_sz = 10
    num_im = 100

    # Create a mock TICA cube that still has the error array
    # Cube with TICA headers to copy to cube with error array
    spoc_cube_error = CubeFactory() 
    spoc_ffi_files = create_test_ffis(img_sz, num_im, dir_name=tmpdir)
    cube_error_file = spoc_cube_error.make_cube(spoc_ffi_files, 
                                                path.join(tmpdir, "out_dir", "test_error_cube.fits"), 
                                                verbose=False)

    # Cube with error array
    tica_cube_no_error = TicaCubeFactory()
    tica_ffi_files = create_test_ffis(img_sz, num_im, dir_name=tmpdir, product='TICA')
    tica_cube_no_error_file = tica_cube_no_error.make_cube(tica_ffi_files, 
                                                           path.join(tmpdir, "out_dir", "test_add_error_cube.fits"), 
                                                           verbose=False)

    # Take ImageHDU from the cube with the error array, write into cube with TICA headers
    with fits.open(tica_cube_no_error_file, mode='update') as hdulist:

        tica_hdr = hdulist[1].header
        error_hdr = fits.getheader(cube_error_file, 1)

        # Overwrite header keywords
        for h in error_hdr.keys():
            tica_hdr[h] = error_hdr[h]

        # Overwrite header data
        hdulist[1].data = fits.getdata(cube_error_file, 1)

        # Close file, write all changes
        hdulist.close()

    # Update variable name to reflect addition of error array to TICA cube
    tica_cube_error_file = tica_cube_no_error_file

    # Verify dimensions of new TICA cube with added error array
    assert np.shape(fits.getdata(tica_cube_error_file, 1)) == (img_sz, img_sz, num_im, 2)

    ffi_type = 'TICA'
    # Test cutouts from TICA cube with error array
    test_cube_cutout(tica_cube_error_file, tica_ffi_files, ffi_type, tmp_path)
    test_header_keywords_quality(tica_cube_error_file, ffi_type, tmp_path)
    test_header_keywords_diffs(tica_cube_error_file, ffi_type, tmp_path)
    test_target_pixel_file(tica_cube_error_file, ffi_type, tmp_path)


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_ffi_cube_header(cube_file, ffi_files, ffi_type):
    """Test FFI headers versus cube headers"""

    ffi_header_keys = list(fits.getheader(ffi_files[0], 0).keys())
    cube_header_keys = list(fits.getheader(cube_file, 0).keys())

    # Lists of expected keys for each product type
    if ffi_type == 'SPOC':
        # CHECKSUM in FFI header, not in cube header
        # SECTOR, DATE, ORIGIN in cube header, not in FFI header
        effi_keys = ['CHECKSUM']
        ecube_keys = ['SECTOR', 'DATE', 'ORIGIN']

    else:
        # CHECKSUM, and image dimensions (NAXIS1, NAXIS2) in FFI header, not in cube header
        # SECTOR, DATE, EXTNAME, ORIGIN in cube header, not in FFI header
        effi_keys = ['CHECKSUM', 'NAXIS1', 'NAXIS2']
        ecube_keys = ['SECTOR', 'DATE', 'EXTNAME', 'ORIGIN']

    for k in effi_keys:
        assert k in ffi_header_keys
        assert k not in cube_header_keys

    for k in ecube_keys:
        assert k not in ffi_header_keys
        assert k in cube_header_keys
    

@pytest.mark.parametrize('ffi_type', ['SPOC', 'TICA'])
def test_exceptions(cube_file, ffi_type):
    """
    Testing various error conditions.
    """

    hdu = fits.open(cube_file)
    cube_table = hdu[2].data

    cutout_maker = CutoutFactory()
     
    # Testing when none of the FFIs have good wcs info
    wcsaxes = 'CTYPE2'
    cube_table[wcsaxes] = 'N/A'
    with pytest.raises(Exception, match='No FFI rows contain valid WCS keywords.') as e:
        cutout_maker._parse_table_info(table_data=cube_table)
        assert e.type is wcs.NoWcsKeywordsFoundError
    cube_table[wcsaxes] = 'DEC--TAN-SIP'

    # Testing when nans are present 
    cutout_maker._parse_table_info(table_data=cube_table)
    wcs_orig = cutout_maker.cube_wcs
    # TICA does not have BARYCORR so we can't use the same 
    # test column for both product types. 
    nan_column = 'BARYCORR' if ffi_type == 'SPOC' else 'DEADC'
    cube_table[nan_column] = np.nan
    cutout_maker._parse_table_info(table_data=cube_table)

    assert wcs_orig.to_header_string() == cutout_maker.cube_wcs.to_header_string()

    hdu.close()

    # Testing various off the cube inputs
    cutout_maker.center_coord = SkyCoord("50.91092264 6.40588255", unit='deg')
    with pytest.raises(Exception, match='Cutout location is not in cube footprint!') as e:
        cutout_maker._get_cutout_limits(np.array([5, 5]))
        assert e.type is InvalidQueryError
         
    cutout_maker.center_coord = SkyCoord("257.91092264 6.40588255", unit='deg')
    with pytest.raises(Exception, match='Cutout location is not in cube footprint!') as e:
        cutout_maker._get_cutout_limits(np.array([5, 5]))
        assert e.type is InvalidQueryError


    # Testing the WCS fitting function
    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_maker.cube_wcs, (100, 100))
    assert distmax.deg < 0.003
    assert sigma < 0.03

    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_maker.cube_wcs, (1, 100))
    assert distmax.deg < 0.003
    assert sigma < 0.03

    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_maker.cube_wcs, (100, 2))
    assert distmax.deg < 0.03
    assert sigma < 0.03

    cutout_maker.center_coord = SkyCoord("256.38994124 4.88986771", unit='deg')
    cutout_maker._get_cutout_limits(np.array([5, 500]))

    hdu = fits.open(cube_file)
    cutout_wcs = cutout_maker._get_full_cutout_wcs(hdu[2].header)
    hdu.close()

    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_wcs, (200, 200))
    assert distmax.deg < 0.004
    assert sigma < 0.2

    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_wcs, (100, 5))
    assert distmax.deg < 0.003
    assert sigma < 0.003

    distmax, sigma = cutout_maker._fit_cutout_wcs(cutout_wcs, (3, 100))
    assert distmax.deg < 0.003
    assert sigma < 0.003


@pytest.mark.parametrize("ffi_type", ["SPOC", "TICA"])
def test_inputs(cube_file, ffi_type, tmp_path, capsys):
    """
    Testing with different user input types/combos. And verbose.
    """

    tmpdir = str(tmp_path)

    cutout_maker = CutoutFactory()

    # Setting up
    coord = "256.88 6.38"

    cutout_size = [5, 3]*u.pixel
    cutout_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, output_path=tmpdir, verbose=True)
    captured = capsys.readouterr()
    assert "Image cutout cube shape: (100, 3, 5)" in captured.out
    assert "Using WCS from row 50 out of 100" in captured.out
    assert "Cutout center coordinate: 256.88,6.38" in captured.out
    assert "5x3" in cutout_file

    cutout_size = [5, 3]*u.arcmin
    cutout_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, output_path=tmpdir, verbose=False)
    assert "14x9" in cutout_file

    cutout_size = [5, 3, 9]*u.pixel
    with pytest.warns(InputWarning):
        cutout_file = cutout_maker.cube_cut(cube_file, coord, cutout_size, ffi_type, output_path=tmpdir, verbose=False)
    assert "5x3" in cutout_file
    assert "x9" not in cutout_file


def test_s3_cube_cut(tmp_path: Path):
    """Does using an S3-hosted TESS cube yield correct results?

    This test implements a spot check which verifies whether a cutout
    for Proxima Cen (Sector 38) obtained from an S3-hosted cube
    file yields results that are identical to those returned
    by the Tesscut service.

    To speed up the test and avoid adding astroquery as a dependency,
    the test uses hard-coded reference values which were obtained
    as follows:

    >>> from astroquery.mast import Tesscut  # doctest: +SKIP
    >>> crd = SkyCoord(217.42893801, -62.67949189, unit="deg")  # doctest: +SKIP
    >>> cut = Tesscut.get_cutouts(crd, size=3, sector=38)  # doctest: +SKIP
    >>> cut[0][1].data.shape  # doctest: +SKIP
    (3705,)
    >>> cut[0][1].data['TIME'][0]  # doctest: +SKIP
    2333.8614060219998
    >>> cut[0][1].data['FLUX'][100][0, 0]  # doctest: +SKIP
    2329.8127
    >>> cut[0][1].data['FLUX_ERR'][200][1, 2]  # doctest: +SKIP
    1.1239403
    >>> cut[0][0].header['CAMERA']  # doctest: +SKIP
    2
    """
    # Test case: Proxima Cen in Sector 38 (Camera 2, CCD 2)
    coord = SkyCoord(217.42893801, -62.67949189, unit="deg", frame="icrs")
    cube_file = "s3://stpubdata/tess/public/mast/tess-s0038-2-2-cube.fits"
    cutout_file = CutoutFactory().cube_cut(cube_file, coord, 3, output_path=str(tmp_path))
    hdulist = fits.open(cutout_file)
    assert hdulist[1].data.shape == (3705,)
    assert np.isclose(hdulist[1].data["TIME"][0], 2333.8614060219998)
    assert np.isclose(hdulist[1].data["FLUX"][100][0, 0], 2329.8127)
    assert np.isclose(hdulist[1].data["FLUX_ERR"][200][1, 2], 1.1239403)
    assert hdulist[0].header["CAMERA"] == 2
    hdulist.close()


@pytest.mark.xfail
def test_s3_tica_cube_cut(tmp_path: Path):
    """Does using an S3-hosted TESS cube yield correct results?

    This test implements a spot check which verifies whether a cutout
    for Sector 27 obtained from an S3-hosted cube
    file yields results that are identical to those returned
    by the Tesscut service.

    To speed up the test and avoid adding astroquery as a dependency,
    the test uses hard-coded reference values which were obtained
    as follows:

    >>> from astroquery.mast import Tesscut  # doctest: +SKIP
    >>> crd = SkyCoord(299.27269, -67.14491, unit="deg")  # doctest: +SKIP
    >>> cut = Tesscut.get_cutouts(coordinates=crd, size=3, sector=27, product='TICA')  # doctest: +SKIP
    >>> cut[0][1].data.shape  # doctest: +SKIP
    (3360,)
    >>> cut[0][1].data['TIME'][0]  # doctest: +SKIP
    2036.281565948625
    >>> cut[0][1].data['FLUX'][500][0, 0]  # doctest: +SKIP
    59313.523
    >>> np.sum(cut[0][1].data['FLUX_ERR'])  # doctest: +SKIP
    0.0
    >>> cut[0][0].header['CAMERA']  # doctest: +SKIP
    2
    >>> cut[0][0].header['CCD']  # doctest: +SKIP
    1

    TODO: Remove test `test_tica_cutout_error` when this test starts working
    """
    # Test case: RR Lyrae star in Sector 27 (Camera 2, CCD 1)
    coord = SkyCoord(299.27269, -67.14491, unit="deg", frame="icrs")
    cube_file = "s3://stpubdata/tess/public/mast/tica/tica-s0027-2-1-cube.fits"
    cutout_file = CutoutFactory().cube_cut(cube_file, coord, 3, output_path=str(tmp_path))
    hdulist = fits.open(cutout_file)
    assert hdulist[1].data.shape == (3360,)
    assert np.isclose(hdulist[1].data["TIME"][0], 2036.281565948625)
    assert np.isclose(hdulist[1].data["FLUX"][500][0, 0], 59313.523)
    assert np.sum(hdulist[1].data["FLUX_ERR"]) == 0.0
    assert hdulist[0].header["CAMERA"] == 2
    assert hdulist[0].header["CCD"] == 1
    hdulist.close()


def test_multithreading(tmp_path):
    tmpdir = str(tmp_path)

    cutout_size = 10
    coord = SkyCoord(217.42893801, -62.67949189, unit="deg", frame="icrs")
    cube_file = "s3://stpubdata/tess/public/mast/tess-s0038-2-2-cube.fits"

    cut_factory = CutoutFactory()

    # 1 thread condition (no threading) used as verification for the different thread conditions
    cutout_no_threads = cut_factory.cube_cut(
        cube_file, coordinates=coord, output_path=tmpdir, verbose=False, cutout_size=cutout_size, threads=1
    )
    data_no_threads = fits.getdata(cutout_no_threads)

    # when threads="auto", number of threads is system dependent: cpu_count + 4, limited to max of 32
    # https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor
    for threads in (4, "auto"):
        cutout_threads = cut_factory.cube_cut(
            cube_file, coordinates=coord, output_path=tmpdir, verbose=False, cutout_size=cutout_size, threads=threads
        )
        data_threads = fits.getdata(cutout_threads)

        for ext_name in ("FLUX", "FLUX_ERR"):
            assert np.array_equal(data_no_threads[ext_name], data_threads[ext_name])




import pathlib
from unittest.mock import MagicMock, patch
import numpy as np
import pytest

import asdf
from astropy.modeling import models
from astropy import coordinates as coord
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_skycoord
from gwcs import wcs
from gwcs import coordinate_frames as cf
from astrocut.asdf_cutouts import get_center_pixel, asdf_cut, _get_cutout, _slice_gwcs, _get_cloud_http


def make_wcs(xsize, ysize, ra=30., dec=45.):
    """ create a fake gwcs object """
    # todo - refine this to better reflect roman wcs

    # create transformations
    # - shift coords so array center is at 0, 0 ; reference pixel
    # - scale pixels to correct angular scale
    # - project coords onto sky with TAN projection
    # - transform center pixel to the input celestial coordinate
    pixelshift = models.Shift(-xsize) & models.Shift(-ysize)
    pixelscale = models.Scale(0.1 / 3600.) & models.Scale(0.1 / 3600.)  # 0.1 arcsec/pixel
    tangent_projection = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(ra, dec, 180.)

    # net transforms pixels to sky
    det2sky = pixelshift | pixelscale | tangent_projection | celestial_rotation

    # define the wcs object
    detector_frame = cf.Frame2D(name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix))
    sky_frame = cf.CelestialFrame(reference_frame=coord.ICRS(), name='world', unit=(u.deg, u.deg))
    return wcs.WCS([(detector_frame, det2sky), (sky_frame, None)])


@pytest.fixture()
def makefake():
    """ fixture factory to make a fake gwcs and dataset """

    def _make_fake(nx, ny, ra, dec, zero=False, asint=False):
        # create the wcs
        wcsobj = make_wcs(nx/2, ny/2, ra=ra, dec=dec)
        wcsobj.bounding_box = ((0, nx), (0, ny))

        # create the data
        if zero:
            data = np.zeros([nx, ny])
        else:
            size = nx * ny
            data = np.arange(size).reshape(nx, ny)

        # make a quantity
        data *= (u.electron / u.second)

        # make integer array
        if asint:
            data = data.astype(int)

        return data, wcsobj

    yield _make_fake


@pytest.fixture()
def fakedata(makefake):
    """ fixture to create fake data and wcs """
    # set up initial parameters
    nx = 1000
    ny = 1000
    ra = 30.
    dec = 45.

    yield makefake(nx, ny, ra, dec)


@pytest.fixture()
def make_file(tmp_path, fakedata):
    """ fixture to create a fake dataset """
    # get the fake data
    data, wcsobj = fakedata

    # create meta
    meta = {'wcs': wcsobj}

    # create and write the asdf file
    tree = {'roman': {'data': data, 'meta': meta}}
    af = asdf.AsdfFile(tree)

    path = tmp_path / "roman"
    path.mkdir(exist_ok=True)
    filename = path / "test_roman.asdf"
    af.write_to(filename)

    yield filename


@pytest.fixture()
def output(tmp_path):
    """ fixture to create the output path """
    def _output_file(ext='fits'):
        # create output fits path
        out = tmp_path / "roman"
        out.mkdir(exist_ok=True, parents=True)
        output_file = out / f"test_output_cutout.{ext}" if ext else "test_output_cutout"
        return output_file
    yield _output_file


def test_get_center_pixel(fakedata):
    """ test we can get the correct center pixel """
    # get the fake data
    __, gwcs = fakedata

    pixel_coordinates, wcs = get_center_pixel(gwcs, 30., 45.)
    assert np.allclose(pixel_coordinates, (np.array(500.), np.array(500.)))
    assert np.allclose(wcs.celestial.wcs.crval, np.array([30., 45.]))


@pytest.mark.parametrize('quantity', [True, False], ids=['quantity', 'array'])
def test_get_cutout(output, fakedata, quantity):
    """ test we can create a cutout """
    output_file = output('fits')

    # get the input wcs
    data, gwcs = fakedata
    skycoord = gwcs(25, 25, with_units=True)
    wcs = WCS(gwcs.to_fits_sip())

    # convert quanity data back to array
    if not quantity:
        data = data.value

    # create cutout
    cutout = _get_cutout(data, skycoord, wcs, size=10, outfile=output_file)

    assert_same_coord(5, 10, cutout, wcs)

    # test output
    with fits.open(output_file) as hdulist:
        data = hdulist[0].data
        assert data.shape == (10, 10)
        assert data[5, 5] == 25025


def test_asdf_cutout(make_file, output):
    """ test we can make a cutout """
    output_file = output('fits')
    # make cutout
    ra, dec = (29.99901792, 44.99930555)
    asdf_cut(make_file, ra, dec, cutout_size=10, output_file=output_file)

    # test output
    with fits.open(output_file) as hdulist:
        data = hdulist[0].data
        assert data.shape == (10, 10)
        assert data[5, 5] == 475476


@pytest.mark.parametrize('suffix', ['fits', 'asdf', None])
def test_write_file(make_file, suffix, output):
    """ test we can write an different file types """
    output_file = output(suffix)

    # make cutout
    ra, dec = (29.99901792, 44.99930555)
    asdf_cut(make_file, ra, dec, cutout_size=10, output_file=output_file)

    # if no suffix provided, check that the default output is fits
    if not suffix:
        output_file += ".fits"

    assert pathlib.Path(output_file).exists()


def test_fail_write_asdf(fakedata, output):
    """ test we fail to write an asdf if no gwcs given """
    with pytest.raises(ValueError, match='The original gwcs object is needed when writing to asdf file.'):
        output_file = output('asdf')
        data, gwcs = fakedata
        skycoord = gwcs(25, 25, with_units=True)
        wcs = WCS(gwcs.to_fits_sip())
        _get_cutout(data, skycoord, wcs, size=10, outfile=output_file)


def test_cutout_nofile(make_file, output):
    """ test we can make a cutout with no file output """
    output_file = output()
    # make cutout
    ra, dec = (29.99901792, 44.99930555)
    cutout = asdf_cut(make_file, ra, dec, cutout_size=10, output_file=output_file, write_file=False)

    assert not pathlib.Path(output_file).exists()
    assert cutout.shape == (10, 10)


def test_cutout_poles(makefake):
    """ test we can make cutouts around poles """
    # make fake zero data around the pole
    ra, dec = 315.0, 89.995
    data, gwcs = makefake(1000, 1000, ra, dec, zero=True)

    # add some values (5x5 array)
    data.value[245:250, 245:250] = 1

    # check central pixel is correct
    ss = gwcs(500, 500)
    assert ss == (ra, dec)

    # set input cutout coord
    cc = coord.SkyCoord(284.702, 89.986, unit=u.degree)
    wcs = WCS(gwcs.to_fits_sip())

    # get cutout
    cutout = _get_cutout(data, cc, wcs, size=50, write_file=False)
    assert_same_coord(5, 10, cutout, wcs)

    # check cutout contains all data
    assert len(np.where(cutout.data.value == 1)[0]) == 25


def test_fail_cutout_outside(fakedata):
    """ test we fail when cutout completely outside range """
    data, gwcs = fakedata
    wcs = WCS(gwcs.to_fits_sip())
    cc = coord.SkyCoord(200.0, 50.0, unit=u.degree)

    with pytest.raises(RuntimeError, match='Could not create 2d cutout.  The requested '
                       'cutout does not overlap with the original image'):
        _get_cutout(data, cc, wcs, size=50, write_file=False)


def assert_same_coord(x, y, cutout, wcs):
    """ assert we get the same sky coordinate from cutout and original wcs """
    cutout_coord = pixel_to_skycoord(x, y, cutout.wcs)
    ox, oy = cutout.to_original_position((x, y))
    orig_coord = pixel_to_skycoord(ox, oy, wcs)
    assert cutout_coord == orig_coord


@pytest.mark.parametrize('asint, fill', [(False, None), (True, -9999)], ids=['fillfloat', 'fillint'])
def test_partial_cutout(makefake, asint, fill):
    """ test we get a partial cutout with nans or fill value """
    ra, dec = 30.0, 45.0
    data, gwcs = makefake(100, 100, ra, dec, asint=asint)

    wcs = WCS(gwcs.to_fits_sip())
    cc = coord.SkyCoord(29.999, 44.998, unit=u.degree)
    cutout = _get_cutout(data, cc, wcs, size=50, write_file=False, fill_value=fill)
    assert cutout.shape == (50, 50)
    if asint:
        assert -9999 in cutout.data
    else:
        assert np.isnan(cutout.data).any()


def test_bad_fill(makefake):
    """ test error is raised on bad fill value """
    ra, dec = 30.0, 45.0
    data, gwcs = makefake(100, 100, ra, dec, asint=True)
    wcs = WCS(gwcs.to_fits_sip())
    cc = coord.SkyCoord(29.999, 44.998, unit=u.degree)
    with pytest.raises(ValueError, match='fill_value is inconsistent with the data type of the input array'):
        _get_cutout(data, cc, wcs, size=50, write_file=False)


def test_cutout_raedge(makefake):
    """ test we can make cutouts around ra=0 """
    # make fake zero data around the ra edge
    ra, dec = 0.0, 10.0
    data, gg = makefake(2000, 2000, ra, dec, zero=True)

    # check central pixel is correct
    ss = gg(1001, 1001)
    assert pytest.approx(ss, abs=1e-3) == (ra, dec)

    # set input cutout coord
    cc = coord.SkyCoord(0.001, 9.999, unit=u.degree)
    wcs = WCS(gg.to_fits_sip())

    # get cutout
    cutout = _get_cutout(data, cc, wcs, size=100, write_file=False)
    assert_same_coord(5, 10, cutout, wcs)

    # assert the RA cutout bounds are > 359 and < 0
    bounds = gg(*cutout.bbox_original, with_units=True)
    assert bounds[0].ra.value > 359
    assert bounds[1].ra.value < 0.1


def test_slice_gwcs(fakedata):
    """ test we can slice a gwcs object """
    data, gwcsobj = fakedata
    skycoord = gwcsobj(250, 250)
    wcs = WCS(gwcsobj.to_fits_sip())

    cutout = _get_cutout(data, skycoord, wcs, size=50, write_file=False)

    sliced = _slice_gwcs(gwcsobj, cutout.slices_original)

    # check coords between slice and original gwcs
    assert cutout.center_cutout == (24.5, 24.5)
    assert sliced.array_shape == (50, 50)
    assert sliced(*cutout.input_position_cutout) == gwcsobj(*cutout.input_position_original)
    assert gwcsobj(*cutout.center_original) == sliced(*cutout.center_cutout)

    # assert same sky footprint between slice and original
    # gwcs footprint/bounding_box expects ((x0, x1), (y0, y1)) but cutout.bbox is in ((y0, y1), (x0, x1))
    assert (gwcsobj.footprint(bounding_box=tuple(reversed(cutout.bbox_original))) == sliced.footprint()).all()


@patch('s3fs.S3FileSystem')
def test_get_cloud_http(mock_s3fs):
    """ test we can get HTTP URI of cloud resource """
    # mock s3 file system operations
    HTTP_URI = "http_test"
    mock_file = MagicMock()
    mock_fs = MagicMock()
    mock_file.url.return_value = HTTP_URI
    mock_fs.open.return_value.__enter__.return_value = mock_file
    mock_s3fs.return_value = mock_fs

    s3_uri = "s3://test_bucket/test_file.asdf"
    http_uri = _get_cloud_http(s3_uri)

    assert http_uri == HTTP_URI
    mock_s3fs.assert_called_once_with()
    mock_fs.open.assert_called_once_with(s3_uri, 'rb')
    mock_file.url.assert_called_once()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module implements the cutout functionality."""

import os
import warnings
from concurrent.futures import ThreadPoolExecutor
from itertools import product
from time import time
from typing import Any, Dict, Literal, Union

import astropy.units as u
import numpy as np
from astropy import wcs
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time

from . import __version__
from .exceptions import InputWarning, InvalidQueryError
from .utils.wcs_fitting import fit_wcs_from_points

# todo: investigate why for small cutouts the astropy version is not working
# from astropy.wcs.utils import fit_wcs_from_points



class CutoutFactory():
    """
    Class for creating image cutouts.

    This class encompasses all of the cutout functionality.  
    In the current version this means creating cutout target pixel files from both 
    SPOC (Science Processing Operations Center) and TICA (Tess Image CAlibration) 
    full frame image cubes.

    Future versions will include more generalized cutout functionality.
    """

    def __init__(self):
        """
        Initialization function.
        """

        self.product = None
        self.cube_wcs = None  # WCS information from the image cube
        self.cutout_wcs = None  # WCS information (linear) for the cutout
        self.cutout_wcs_fit = {
            "WCS_MSEP": [None, "[deg] Max offset between cutout WCS and FFI WCS"],
            "WCS_SIG": [None, "[deg] Error measurement of cutout WCS fit"],
        }

        self.cutout_lims = np.zeros((2, 2), dtype=int)  # Cutout pixel limits, [[ymin,ymax],[xmin,xmax]]
        self.center_coord = None  # Central skycoord

        # Extra keywords from the FFI image headers in SPOC.
        # These are applied to both SPOC and TICA cutouts for consistency.
        self.img_kwds = {
            "BACKAPP": [None, "background is subtracted"],
            "CDPP0_5": [None, "RMS CDPP on 0.5-hr time scales"],
            "CDPP1_0": [None, "RMS CDPP on 1.0-hr time scales"],
            "CDPP2_0": [None, "RMS CDPP on 2.0-hr time scales"],
            "CROWDSAP": [None, "Ratio of target flux to total flux in op. ap."],
            "DEADAPP": [None, "deadtime applied"],
            "DEADC": [None, "deadtime correction"],
            "EXPOSURE": [None, "[d] time on source"],
            "FLFRCSAP": [None, "Frac. of target flux w/in the op. aperture"],
            "FRAMETIM": [None, "[s] frame time [INT_TIME + READTIME]"],
            "FXDOFF": [None, "compression fixed offset"],
            "GAINA": [None, "[electrons/count] CCD output A gain"],
            "GAINB": [None, "[electrons/count] CCD output B gain"],
            "GAINC": [None, "[electrons/count] CCD output C gain"],
            "GAIND": [None, "[electrons/count] CCD output D gain"],
            "INT_TIME": [None, "[s] photon accumulation time per frame"],
            "LIVETIME": [None, "[d] TELAPSE multiplied by DEADC"],
            "MEANBLCA": [None, "[count] FSW mean black level CCD output A"],
            "MEANBLCB": [None, "[count] FSW mean black level CCD output B"],
            "MEANBLCC": [None, "[count] FSW mean black level CCD output C"],
            "MEANBLCD": [None, "[count] FSW mean black level CCD output D"],
            "NREADOUT": [None, "number of read per cadence"],
            "NUM_FRM": [None, "number of frames per time stamp"],
            "READNOIA": [None, "[electrons] read noise CCD output A"],
            "READNOIB": [None, "[electrons] read noise CCD output B"],
            "READNOIC": [None, "[electrons] read noise CCD output C"],
            "READNOID": [None, "[electrons] read noise CCD output D"],
            "READTIME": [None, "[s] readout time per frame"],
            "TIERRELA": [None, "[d] relative time error"],
            "TIMEDEL": [None, "[d] time resolution of data"],
            "TIMEPIXR": [None, "bin time beginning=0 middle=0.5 end=1"],
            "TMOFST11": [None, "(s) readout delay for camera 1 and ccd 1"],
            "VIGNAPP": [None, "vignetting or collimator correction applied"],
        }

    def _parse_table_info(self, table_data, verbose=False):
        """
        Takes the header and the middle entry from the cube table (EXT 2) of image header data,
        builds a WCS object that encapsulates the given WCS information,
        and collects into a dictionary the other keywords we care about.

        The WCS is stored in ``self.cube_wcs``, and the extra keywords in ``self.img_kwds``

        Parameters
        ----------
        table_data : `~astropy.io.fits.fitsrec.FITS_rec`
            The cube image header data table.
        """

        # Populating `table_row` with the primary header keywords 
        # of the middle FFI
        data_ind = len(table_data)//2
        table_row = None

        while table_row is None:
            table_row = table_data[data_ind]

            # Making sure we have a row with wcs info.
            wcsaxes_keyword = 'CTYPE2'
            if table_row[wcsaxes_keyword] != 'DEC--TAN-SIP':
                table_row = None
                data_ind += 1
                if data_ind == len(table_data):
                    raise wcs.NoWcsKeywordsFoundError("No FFI rows contain valid WCS keywords.")

        if verbose:
            print("Using WCS from row {} out of {}".format(data_ind, len(table_data)))

        # Turning the table row into a new header object
        wcs_header = fits.header.Header()
        for col in table_data.columns:
            
            wcs_val = table_row[col.name]
            if (not isinstance(wcs_val, str)) and (np.isnan(wcs_val)):
                continue  # Just skip nans

            wcs_header[col.name] = wcs_val
            
        # Setting the cube wcs
        self.cube_wcs = wcs.WCS(wcs_header, relax=True)

        # Filling the img_kwds dictionary while we are here
        for kwd in self.img_kwds:
            self.img_kwds[kwd][0] = wcs_header.get(kwd)
        # Adding the info about which FFI we got the WCS info from
        self.img_kwds["WCS_FFI"] = [table_data[data_ind]["FFI_FILE"],
                                    "FFI used for cutout WCS"]

            
    def _get_cutout_limits(self, cutout_size):
        """
        Takes the center coordinates, cutout size, and the wcs from
        which the cutout is being taken and returns the x and y pixel limits
        for the cutout.

        Parameters
        ----------
        cutout_size : array
            [ny,nx] in with ints (pixels) or astropy quantities

        Returns
        -------
        response : `numpy.array`
            The cutout pixel limits in an array of the form [[ymin,ymax],[xmin,xmax]]
        """
        
        # Note: This is returning the center pixel in 1-up
        try:
            center_pixel = self.center_coord.to_pixel(self.cube_wcs, 1)
        except wcs.NoConvergence:  # If wcs can't converge, center coordinate is far from the footprint
            raise InvalidQueryError("Cutout location is not in cube footprint!")

        lims = np.zeros((2, 2), dtype=int)

        for axis, size in enumerate(cutout_size):

            if not isinstance(size, u.Quantity):  # assume pixels
                dim = size / 2
            elif size.unit == u.pixel:  # also pixels
                dim = size.value / 2
            elif size.unit.physical_type == 'angle':
                pixel_scale = u.Quantity(wcs.utils.proj_plane_pixel_scales(self.cube_wcs)[axis],
                                         self.cube_wcs.wcs.cunit[axis])
                dim = (size / pixel_scale).decompose() / 2

            lims[axis, 0] = int(np.round(center_pixel[axis] - 1 - dim))
            lims[axis, 1] = int(np.round(center_pixel[axis] - 1 + dim))

            # The case where the requested area is so small it rounds to zero
            if lims[axis, 0] == lims[axis, 1]:
                lims[axis, 0] = int(np.floor(center_pixel[axis] - 1))
                lims[axis, 1] = int(np.ceil(center_pixel[axis] - 1))

        # Checking at least some of the cutout is on the cube
        if ((lims[0, 0] <= 0) and (lims[0, 1] <= 0)) or ((lims[1, 0] <= 0) and (lims[1, 1] <= 0)):
            raise InvalidQueryError("Cutout location is not in cube footprint!")

        self.cutout_lims = lims


    def _get_full_cutout_wcs(self, cube_table_header):
        """
        Starting with the full FFI WCS and adjusting it for the cutout WCS.
        Adjusts CRPIX values and adds physical WCS keywords.

        Parameters
        ----------
        cube_table_header :  `~astropy.io.fits.Header`
           The FFI cube header for the data table extension. This allows the cutout WCS information
           to more closely match the mission TPF format.

        Resturns
        --------
        response :  `~astropy.wcs.WCS`
            The cutout WCS object including SIP distortions.
        """
        
        wcs_header = self.cube_wcs.to_header(relax=True)

        # Using table comment rather than the default ones if available
        for kwd in wcs_header:
            if cube_table_header.get(kwd, None):
                wcs_header.comments[kwd] = cube_table_header[kwd]

        # Adjusting the CRPIX values
        wcs_header["CRPIX1"] -= self.cutout_lims[0, 0]
        wcs_header["CRPIX2"] -= self.cutout_lims[1, 0]

        # Adding the physical wcs keywords
        wcs_header.set("WCSNAMEP", "PHYSICAL", "name of world coordinate system alternate P")
        wcs_header.set("WCSAXESP", 2, "number of WCS physical axes")
    
        wcs_header.set("CTYPE1P", "RAWX", "physical WCS axis 1 type CCD col")
        wcs_header.set("CUNIT1P", "PIXEL", "physical WCS axis 1 unit")
        wcs_header.set("CRPIX1P", 1, "reference CCD column")
        wcs_header.set("CRVAL1P", self.cutout_lims[0, 0] + 1, "value at reference CCD column")
        wcs_header.set("CDELT1P", 1.0, "physical WCS axis 1 step")
                
        wcs_header.set("CTYPE2P", "RAWY", "physical WCS axis 2 type CCD col")
        wcs_header.set("CUNIT2P", "PIXEL", "physical WCS axis 2 unit")
        wcs_header.set("CRPIX2P", 1, "reference CCD row")
        wcs_header.set("CRVAL2P", self.cutout_lims[1, 0] + 1, "value at reference CCD row")
        wcs_header.set("CDELT2P", 1.0, "physical WCS axis 2 step")

        return wcs.WCS(wcs_header)

    
    def _fit_cutout_wcs(self, cutout_wcs, cutout_shape):
        """
        Given a full (including SIP coefficients) wcs for the cutout, 
        calculate the best fit linear wcs, and a measure of the goodness-of-fit.
        
        The new WCS is stored in ``self.cutout_wcs``.
        Goodness-of-fit measures are returned and stored in ``self.cutout_wcs_fit``.

        Parameters
        ----------
        cutout_wcs :  `~astropy.wcs.WCS`
            The full (including SIP coefficients) cutout WCS object 
        cutout_shape : tuple
            The shape of the cutout in the form (width, height).

        Returns
        -------
        response : tuple
            Goodness-of-fit statistics. (max dist, sigma)
        """

        # Getting matched pixel, world coordinate pairs
        # We will choose no more than 100 pixels spread evenly throughout the image
        # Centered on the center pixel.
        # To do this we the appropriate "step size" between pixel coordinates
        # (i.e. we take every ith pixel in each row/column) [TODOOOOOO]
        # For example in a 5x7 cutout with i = 2 we get:
        #
        # xxoxoxx
        # ooooooo
        # xxoxoxx
        # ooooooo
        # xxoxoxx
        #
        # Where x denotes the indexes that will be used in the fit.
        y, x = cutout_shape
        i = 1
        while (x/i)*(y/i) > 100:
            i += 1
            
        xvals = list(reversed(range(x//2, -1, -i)))[:-1] + list(range(x//2, x, i))
        if xvals[-1] != x-1:
            xvals += [x-1]
        if xvals[0] != 0:
            xvals = [0] + xvals
        
        yvals = list(reversed(range(y//2, -1, -i)))[:-1] + list(range(y//2, y, i))
        if yvals[-1] != y-1:
            yvals += [y-1]
        if yvals[0] != 0:
            yvals = [0] + yvals
        
        pix_inds = np.array(list(product(xvals, yvals)))
        world_pix = SkyCoord(cutout_wcs.all_pix2world(pix_inds, 0), unit='deg')

        # Getting the fit WCS
        linear_wcs = fit_wcs_from_points([pix_inds[:, 0], pix_inds[:, 1]], world_pix, proj_point='center')

        self.cutout_wcs = linear_wcs

        # Checking the fit (we want to use all of the pixels for this)
        pix_inds = np.array(list(product(list(range(cutout_shape[1])), list(range(cutout_shape[0])))))
        world_pix = SkyCoord(cutout_wcs.all_pix2world(pix_inds, 0), unit='deg')
        world_pix_new = SkyCoord(linear_wcs.all_pix2world(pix_inds, 0), unit='deg')

        dists = world_pix.separation(world_pix_new).to('deg')
        sigma = np.sqrt(sum(dists.value**2))

        self.cutout_wcs_fit['WCS_MSEP'][0] = dists.max().value
        self.cutout_wcs_fit['WCS_SIG'][0] = sigma

        return (dists.max(), sigma)
    
  
    def _get_cutout_wcs_dict(self):
        """
        Transform the cutout WCS object into the cutout column WCS keywords.
        Adds the physical keywords for transformation back from cutout to location on FFI.
        This is a very TESS specific function.
        
        Returns
        -------
        response: dict
            Cutout wcs column header keywords as dictionary of 
            ``{<kwd format string>: [value, desc]} pairs.``
        """
        wcs_header = self.cutout_wcs.to_header()

        cutout_wcs_dict = dict()

        ## Cutout array keywords ##

        cutout_wcs_dict["WCAX{}"] = [wcs_header['WCSAXES'], "number of WCS axes"]
        # TODO: check for 2? this must be two

        cutout_wcs_dict["1CTYP{}"] = [wcs_header["CTYPE1"], "right ascension coordinate type"]
        cutout_wcs_dict["2CTYP{}"] = [wcs_header["CTYPE2"], "declination coordinate type"]
        
        cutout_wcs_dict["1CRPX{}"] = [wcs_header["CRPIX1"], "[pixel] reference pixel along image axis 1"]
        cutout_wcs_dict["2CRPX{}"] = [wcs_header["CRPIX2"], "[pixel] reference pixel along image axis 2"]
    
        cutout_wcs_dict["1CRVL{}"] = [wcs_header["CRVAL1"], "[deg] right ascension at reference pixel"]
        cutout_wcs_dict["2CRVL{}"] = [wcs_header["CRVAL2"], "[deg] declination at reference pixel"]

        cutout_wcs_dict["1CUNI{}"] = [wcs_header["CUNIT1"], "physical unit in column dimension"]
        cutout_wcs_dict["2CUNI{}"] = [wcs_header["CUNIT2"], "physical unit in row dimension"]

        cutout_wcs_dict["1CDLT{}"] = [wcs_header["CDELT1"], "[deg] pixel scale in RA dimension"]
        cutout_wcs_dict["2CDLT{}"] = [wcs_header["CDELT1"], "[deg] pixel scale in DEC dimension"]

        cutout_wcs_dict["11PC{}"] = [wcs_header["PC1_1"], "Coordinate transformation matrix element"]
        cutout_wcs_dict["12PC{}"] = [wcs_header["PC1_2"], "Coordinate transformation matrix element"]
        cutout_wcs_dict["21PC{}"] = [wcs_header["PC2_1"], "Coordinate transformation matrix element"]
        cutout_wcs_dict["22PC{}"] = [wcs_header["PC2_2"], "Coordinate transformation matrix element"]

        
        ## Physical keywords ##
        # TODO: Make sure these are correct
        cutout_wcs_dict["WCSN{}P"] = ["PHYSICAL", "table column WCS name"]
        cutout_wcs_dict["WCAX{}P"] = [2, "table column physical WCS dimensions"]
    
        cutout_wcs_dict["1CTY{}P"] = ["RAWX", "table column physical WCS axis 1 type, CCD col"]
        cutout_wcs_dict["2CTY{}P"] = ["RAWY", "table column physical WCS axis 2 type, CCD row"]
    
        cutout_wcs_dict["1CUN{}P"] = ["PIXEL", "table column physical WCS axis 1 unit"]
        cutout_wcs_dict["2CUN{}P"] = ["PIXEL", "table column physical WCS axis 2 unit"]
    
        cutout_wcs_dict["1CRV{}P"] = [self.cutout_lims[0, 0] + 1,
                                      "table column physical WCS ax 1 ref value"]
        cutout_wcs_dict["2CRV{}P"] = [self.cutout_lims[1, 0] + 1,
                                      "table column physical WCS ax 2 ref value"]

        # TODO: can we calculate these? or are they fixed?
        cutout_wcs_dict["1CDL{}P"] = [1.0, "table column physical WCS a1 step"]    
        cutout_wcs_dict["2CDL{}P"] = [1.0, "table column physical WCS a2 step"]
    
        cutout_wcs_dict["1CRP{}P"] = [1, "table column physical WCS a1 reference"]
        cutout_wcs_dict["2CRP{}P"] = [1, "table column physical WCS a2 reference"]

        return cutout_wcs_dict

    def _get_cutout(self, transposed_cube, threads: Union[int, Literal["auto"]] = 1, verbose=True):
        """
        Making a cutout from an image/uncertainty cube that has been transposed
        to have time on the longest axis.

        Parameters
        ----------
        transposed_cube : `numpy.array`
            Transposed image/uncertainty array.
        threads : int, "auto", default=1
            Number of threads to use when making remote (e.g. s3) cutouts, will not use threads for local access
            <=1 disables the threadpool, >1 sets threadpool to the specified number of threads,
            "auto" uses `concurrent.futures.ThreadPoolExecutor`'s default: cpu_count + 4, limit to max of 32
        verbose :  bool
            Optional. If true intermediate information is printed.

        Returns
        -------
        response :  `numpy.array`, `numpy.array` (or `None`), `numpy.array`
            The untransposed image cutout array,
            the untransposed uncertainty cutout array (if `self.product` is SPOC, otherwise returns `None`),
            and the aperture array (an array the size of a single cutout
            that is 1 where there is image data and 0 where there isn't)
        """

        # These limits are not guarenteed to be within the image footprint
        xmin, xmax = self.cutout_lims[1]
        ymin, ymax = self.cutout_lims[0]

        # Get the image array limits
        xmax_cube, ymax_cube, _, _ = transposed_cube.shape

        # Adjust limits and figuring out the padding
        padding = np.zeros((3, 2), dtype=int)
        if xmin < 0:
            padding[1, 0] = -xmin
            xmin = 0
        if ymin < 0:
            padding[2, 0] = -ymin
            ymin = 0
        if xmax > xmax_cube:
            padding[1, 1] = xmax - xmax_cube
            xmax = xmax_cube
        if ymax > ymax_cube:
            padding[2, 1] = ymax - ymax_cube
            ymax = ymax_cube       
        
        # Doing the cutout
        if threads == "auto" or threads > 1:
            max_workers = None if threads == "auto" else threads
            with ThreadPoolExecutor(max_workers=max_workers) as pool:
                # increase download performance by making remote cutouts inside of a threadpool
                # astropy.io.fits Section class executes a list comprehension, generating a sequence of http range
                # requests (through fsspec) one per our slowest moving dimension (x).
                # https://github.com/astropy/astropy/blob/0a71105fbaa71439206d496a44df11abc0e3ac96/astropy/io/fits/hdu/image.py#L1059
                # By running inside a threadpool, these requests instead execute concurrently which increases cutout
                # throughput. Extremely small cutouts (< 4px in x dim) are not likely to see an improvement
                cutouts = list(pool.map(lambda x: transposed_cube[x, ymin:ymax, :, :], range(xmin, xmax)))
            # stack the list of cutouts
            cutout = np.stack(cutouts)
        else:
            cutout = transposed_cube[xmin:xmax, ymin:ymax, :, :]

        img_cutout = cutout[:, :, :, 0].transpose((2, 0, 1))
        if self.product == "SPOC":
            uncert_cutout = cutout[:, :, :, 1].transpose((2, 0, 1))
        else:
            uncert_cutout = None
        # Making the aperture array
        aperture = np.ones((xmax-xmin, ymax-ymin), dtype=np.int32)

        # Adding padding to the cutouts so that it's the expected size
        if padding.any():  # only do if we need to pad
            img_cutout = np.pad(img_cutout, padding, 'constant', constant_values=np.nan)
            if self.product == "SPOC":
                uncert_cutout = np.pad(uncert_cutout, padding, 'constant', constant_values=np.nan)
            aperture = np.pad(aperture, padding[1:], 'constant', constant_values=0)

        if verbose:
            print("Image cutout cube shape: {}".format(img_cutout.shape))
            if self.product == "SPOC":
                print("Uncertainty cutout cube shape: {}".format(uncert_cutout.shape))
    
        return img_cutout, uncert_cutout, aperture


    def _update_primary_header(self, primary_header):
        """
        Updates the primary header for the cutout target pixel file by filling in 
        the object ra and dec with the central cutout coordinates and filling in
        the rest of the TESS target pixel file keywords wil 0/empty strings
        as we do not have access to this information.
        This is a TESS-specific function.

        Parameter(s)
        ----------
        primary_header : `~astropy.io.fits.Header`
            The primary header from the cube file that will be modified in place for the cutout.
        """

        # Adding cutout specific headers
        primary_header['CREATOR'] = ('astrocut', 'software used to produce this file')
        primary_header['PROCVER'] = (__version__, 'software version')
        primary_header['FFI_TYPE'] = (self.product, 'the FFI type used to make the cutouts')
        # TODO : The name of FIRST_FFI (and LAST_FFI) is too long to be a header kwd value.
        # Find a way to include these in the headers without breaking astropy (maybe abbreviate?).
        # primary_header['FIRST_FFI'] = (self.first_ffi, 'the FFI used for the primary header 
        # keyword values, except TSTOP')
        # primary_header['LAST_FFI'] = (self.last_ffi, 'the FFI used for the TSTOP keyword value')

        primary_header['RA_OBJ'] = (self.center_coord.ra.deg, '[deg] right ascension')
        primary_header['DEC_OBJ'] = (self.center_coord.dec.deg, '[deg] declination')

        timeref = 'SOLARSYSTEM' if self.product == 'SPOC' else None
        tassign = 'SPACECRAFT' if self.product == 'SPOC' else None
        primary_header['TIMEREF'] = (timeref, 'barycentric correction applied to times')        
        primary_header['TASSIGN'] = (tassign, 'where time is assigned')
        primary_header['TIMESYS'] = ('TDB', 'time system is Barycentric Dynamical Time (TDB)')

        primary_header['BJDREFI'] = (2457000, 'integer part of BTJD reference date')           
        primary_header['BJDREFF'] = (0.00000000, 'fraction of the day in BTJD reference date')    
        primary_header['TIMEUNIT'] = ('d', 'time unit for TIME, TSTART and TSTOP')

        delete_kwds_wildcards = ['SC_*', 'RMS*', 'A_*', 'AP_*', 'B_*', 'BP*', 'GAIN*', 'TESS_*', 'CD*',
                                 'CT*', 'CRPIX*', 'CRVAL*', 'MJD*']
        delete_kwds = ['COMMENT', 'FILTER', 'TIME', 'EXPTIME', 'ACS_MODE', 'DEC_TARG', 'FLXWIN', 'RA_TARG',
                       'CCDNUM', 'CAMNUM', 'WCSGDF', 'UNITS', 'CADENCE', 'SCIPIXS', 'INT_TIME', 'PIX_CAT',
                       'REQUANT', 'DIFF_HUF', 'PRIM_HUF', 'QUAL_BIT', 'SPM', 'STARTTJD', 'ENDTJD', 'CRM',
                       'TJD_ZERO', 'CRM_N', 'ORBIT_ID', 'MIDTJD']

        if self.product == 'TICA':

            # Adding some missing kwds not in TICA (but in Ames-produced SPOC ffis)
            primary_header['EXTVER'] = ('1', 'extension version number (not format version)')
            primary_header['SIMDATA'] = (False, 'file is based on simulated data')
            primary_header['NEXTEND'] = ('2', 'number of standard extensions')
            primary_header['TSTART'] = (primary_header['STARTTJD'], 'observation start time in TJD of first FFI')
            primary_header['TSTOP'] = (primary_header['ENDTJD'], 'observation stop time in TJD of last FFI')
            primary_header['CAMERA'] = (primary_header['CAMNUM'], 'Camera number')
            primary_header['CCD'] = (primary_header['CCDNUM'], 'CCD chip number')
            primary_header['ASTATE'] = (None, 'archive state F indicates single orbit processing')
            primary_header['CRMITEN'] = (primary_header['CRM'], 'spacecraft cosmic ray mitigation enabled')
            primary_header['CRBLKSZ'] = (None, '[exposures] s/c cosmic ray mitigation block siz')
            primary_header['FFIINDEX'] = (primary_header['CADENCE'], 'number of FFI cadence interval of first FFI')
            primary_header['DATA_REL'] = (None, 'data release version number')

            date_obs = Time(primary_header['TSTART']+primary_header['BJDREFI'], format='jd').iso
            date_end = Time(primary_header['TSTOP']+primary_header['BJDREFI'], format='jd').iso
            primary_header['DATE-OBS'] = (date_obs, 'TSTART as UTC calendar date of first FFI')
            primary_header['DATE-END'] = (date_end, 'TSTOP as UTC calendar date of last FFI')

            primary_header['FILEVER'] = (None, 'file format version')
            primary_header['RADESYS'] = (None, 'reference frame of celestial coordinates')
            primary_header['SCCONFIG'] = (None, 'spacecraft configuration ID')
            primary_header['TIMVERSN'] = (None, 'OGIP memo number for file format')

            # Bulk removal with wildcards. Most of these should only live in EXT 1 header.
            for kwd in delete_kwds_wildcards:
                try:
                    del primary_header[kwd]
                except KeyError:
                    continue

            # Removal of specific kwds not relevant for cutouts.
            # Most likely these describe a single FFI, and not
            # the whole cube, which is misleading because we are 
            # working with entire stacks of FFIs. Other keywords 
            # are analogs to ones that have already been added 
            # to the primary header in the lines above.
            for kwd in delete_kwds:
                try:
                    del primary_header[kwd]
                except KeyError:
                    continue

        telapse = primary_header.get("TSTOP", 0) - primary_header.get("TSTART", 0)
        primary_header['TELAPSE '] = (telapse, '[d] TSTOP - TSTART')

        # Updating card comment to be more explicit
        primary_header['DATE'] = (primary_header['DATE'], 'FFI cube creation date')

        # Specifying that some of these headers keyword values are inherited from the first FFI
        if self.product == 'SPOC':
            primary_header['TSTART'] = (primary_header['TSTART'], 'observation start time in TJD of first FFI')
            primary_header['TSTOP'] = (primary_header['TSTOP'], 'observation stop time in TJD of last FFI')
            primary_header['DATE-OBS'] = (primary_header['DATE-OBS'], 'TSTART as UTC calendar date of first FFI')
            primary_header['DATE-END'] = (primary_header['DATE-END'], 'TSTOP as UTC calendar date of last FFI')
            primary_header['FFIINDEX'] = (primary_header['FFIINDEX'], 'number of FFI cadence interval of first FFI')

        # These are all the things in the TESS pipeline tpfs about the object that we can't fill
        primary_header['OBJECT'] = ("", 'string version of target id ')
        primary_header['TCID'] = (0, 'unique tess target identifier')
        primary_header['PXTABLE'] = (0, 'pixel table id') 
        primary_header['PMRA'] = (0.0, '[mas/yr] RA proper motion') 
        primary_header['PMDEC'] = (0.0, '[mas/yr] Dec proper motion') 
        primary_header['PMTOTAL'] = (0.0, '[mas/yr] total proper motion') 
        primary_header['TESSMAG'] = (0.0, '[mag] TESS magnitude') 
        primary_header['TEFF'] = (0.0, '[K] Effective temperature') 
        primary_header['LOGG'] = (0.0, '[cm/s2] log10 surface gravity') 
        primary_header['MH'] = (0.0, '[log10([M/H])] metallicity') 
        primary_header['RADIUS'] = (0.0, '[solar radii] stellar radius')
        primary_header['TICVER'] = (0, 'TICVER')
        primary_header['TICID'] = (None, 'unique tess target identifier')

    
    def _add_column_wcs(self, table_header, wcs_dict):
        """
        Adds WCS information for the array columns to the cutout table header.

        Parameters
        ----------
        table_header : `~astropy.io.fits.Header`
            The table header for the cutout table that will be modified in place to include 
            WCS information.
        wcs_dict : dict
            Dictionary of wcs keyword/value pairs to be added to each array column in the 
            cutout table header.
        """

        wcs_col = False  # says if column is one that requires wcs info
    
        for kwd in table_header:

            # Adding header descriptions for the table keywords
            if "TTYPE" in kwd:
                table_header.comments[kwd] = "column name"
            elif "TFORM" in kwd:
                table_header.comments[kwd] = "column format"
            elif "TUNIT" in kwd:
                table_header.comments[kwd] = "unit"
            elif "TDISP" in kwd:
                table_header.comments[kwd] = "display format"
            elif "TDIM" in kwd:
                table_header.comments[kwd] = "multi-dimensional array spec"
                wcs_col = True  # if column holds 2D array need to add wcs info
            elif "TNULL" in kwd:
                table_header.comments[kwd] = "null value"

            # Adding wcs info if necessary
            if (kwd[:-1] == "TTYPE") and wcs_col:
                wcs_col = False  # reset
                for wcs_key, (val, com) in wcs_dict.items():
                    table_header.insert(kwd, (wcs_key.format(int(kwd[-1])-1), val, com))

        
    def _add_img_kwds(self, table_header):
        """
        Adding extra keywords to the table header.

        Parameters
        ----------
        table_header : `~astropy.io.fits.Header`
            The table header to add keywords to.  It will be modified in place.
        """

        for key in self.img_kwds:
            # We'll skip these TICA-specific image keywords that are single-FFI specific
            # or just not helpful
            if (key == 'TIME') or (key == 'EXPTIME') or (key == 'FILTER'):
                continue
            table_header[key] = tuple(self.img_kwds[key])

            
    def _apply_header_inherit(self, hdu_list):
        """
        The INHERIT keyword indicated that keywords from the primary header should be duplicated in 
        the headers of all subsequent extensions.  This function performs this addition in place to 
        the given hdu list.
        
        Parameters
        ----------
        hdu_list : `~astropy.io.fits.HDUList`
            The hdu list to aaply the INHERIT keyword to.
        """
    
        primary_header = hdu_list[0].header

        reserved_kwds = ["COMMENT", "SIMPLE", "BITPIX", "EXTEND", "NEXTEND"]

        for hdu in hdu_list[1:]:
            if hdu.header.get("INHERIT", False):
                for kwd in primary_header:
                    if (kwd not in hdu.header) and (kwd not in reserved_kwds):
                        hdu.header[kwd] = (primary_header[kwd], primary_header.comments[kwd])
            

    def _build_tpf(self, cube_fits, img_cube, uncert_cube, cutout_wcs_dict, aperture):
        """
        Building the cutout target pixel file (TPF) and formatting it to match TESS pipeline TPFs.

        Paramters
        ---------
        cube_fits : `~astropy.io.fits.hdu.hdulist.HDUList`
            The cube hdu list.
        img_cube : `numpy.array`
            The untransposed image cutout array
        uncert_cube : `numpy.array`
            The untransposed uncertainty cutout array.
            This value is set to `None` by default for TICA cutouts.
        cutout_wcs_dict : dict
            Dictionary of wcs keyword/value pairs to be added to each array 
            column in the cutout table header.
        aperture : `numpy.array`
            The aperture array (an array the size of a single cutout 
            that is 1 where there is image data and 0 where there isn't)        
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response :  `~astropy.io.fits.HDUList`
            Target pixel file HDU list
        """
        
        # The primary hdu is just the main header, which is the same
        # as the one on the cube file
        primary_hdu = cube_fits[0]
        self._update_primary_header(primary_hdu.header)

        cols = list()

        # Adding the cutouts
        tform = str(img_cube[0].size) + "E"
        dims = str(img_cube[0].shape[::-1])
        empty_arr = np.zeros(img_cube.shape)
        # Adding the Time relates columns
        start = 'TSTART' if self.product == 'SPOC' else 'STARTTJD'
        stop = 'TSTOP' if self.product == 'SPOC' else 'ENDTJD'
        cols.append(fits.Column(name='TIME', format='D', unit='BJD - 2457000, days', disp='D14.7',
                                array=(cube_fits[2].columns[start].array + cube_fits[2].columns[stop].array)/2))

        if self.product == 'SPOC':
            cols.append(fits.Column(name='TIMECORR', format='E', unit='d', disp='E14.7',
                                    array=cube_fits[2].columns['BARYCORR'].array))

        # Adding CADENCENO as zeros for SPOC b/c we don't have this info
        cadence_array = np.zeros(img_cube.shape)[:, 0, 0] if self.product == 'SPOC'\
            else cube_fits[2].columns['CADENCE'].array
        cols.append(fits.Column(name='CADENCENO', format='J', disp='I10', array=cadence_array))

        # Adding counts (-1 b/c we don't have data)
        cols.append(fits.Column(name='RAW_CNTS', format=tform.replace('E', 'J'), unit='count', dim=dims, disp='I8',
                                array=empty_arr-1, null=-1))

        # Adding flux and flux_err (data we actually have!)
        pixel_unit = 'e-/s' if self.product == 'SPOC' else 'e-'
        flux_err_array = uncert_cube if self.product == 'SPOC' else empty_arr
        cols.append(fits.Column(name='FLUX', format=tform, dim=dims, unit=pixel_unit, disp='E14.7', array=img_cube))
        cols.append(fits.Column(name='FLUX_ERR', format=tform, dim=dims, unit=pixel_unit, disp='E14.7',
                                array=flux_err_array)) 
   
        # Adding the background info (zeros b.c we don't have this info)
        cols.append(fits.Column(name='FLUX_BKG', format=tform, dim=dims, unit=pixel_unit, disp='E14.7',
                                array=empty_arr))
        cols.append(fits.Column(name='FLUX_BKG_ERR', format=tform, dim=dims,
                                unit=pixel_unit, disp='E14.7', array=empty_arr))

        # Adding the quality flags
        data_quality = 'DQUALITY' if self.product == 'SPOC' else 'QUAL_BIT'
        cols.append(fits.Column(name='QUALITY', format='J', disp='B16.16',
                                array=cube_fits[2].columns[data_quality].array))

        # Adding the position correction info (zeros b.c we don't have this info)
        cols.append(fits.Column(name='POS_CORR1', format='E', unit='pixel', disp='E14.7', array=empty_arr[:, 0, 0]))
        cols.append(fits.Column(name='POS_CORR2', format='E', unit='pixel', disp='E14.7', array=empty_arr[:, 0, 0]))

        # Adding the FFI_FILE column (not in the pipeline tpfs)
        cols.append(fits.Column(name='FFI_FILE', format='38A', unit='pixel',
                                array=cube_fits[2].columns['FFI_FILE'].array))
        
        # making the table HDU
        table_hdu = fits.BinTableHDU.from_columns(cols)
        table_hdu.header['EXTNAME'] = 'PIXELS'
        table_hdu.header['INHERIT'] = True
    
        # Adding the wcs keywords to the columns and removing from the header
        self._add_column_wcs(table_hdu.header, cutout_wcs_dict)

        # Adding the extra image keywords
        self._add_img_kwds(table_hdu.header)

        # Building the aperture HDU
        aperture_hdu = fits.ImageHDU(data=aperture)
        aperture_hdu.header['EXTNAME'] = 'APERTURE'
        for kwd, val, cmt in self.cutout_wcs.to_header().cards: 
            aperture_hdu.header.set(kwd, val, cmt)

        # Adding extra aperture keywords (TESS specific)
        aperture_hdu.header.set("NPIXMISS", None, "Number of op. aperture pixels not collected")
        aperture_hdu.header.set("NPIXSAP", None, "Number of pixels in optimal aperture")

        # Adding goodness-of-fit keywords
        for key in self.cutout_wcs_fit:
            aperture_hdu.header[key] = tuple(self.cutout_wcs_fit[key])
        
        aperture_hdu.header['INHERIT'] = True
    
        cutout_hdu_list = fits.HDUList([primary_hdu, table_hdu, aperture_hdu])
        
        self._apply_header_inherit(cutout_hdu_list)

        return cutout_hdu_list

    def cube_cut(
        self,
        cube_file,
        coordinates,
        cutout_size,
        product="SPOC",
        target_pixel_file=None,
        output_path=".",
        threads: Union[int, Literal["auto"]] = 1,
        verbose=False,
    ):
        """
        Takes a cube file (as created by `~astrocut.CubeFactory`), and makes a cutout target pixel
        file of the given size around the given coordinates. The target pixel file is formatted like
        a TESS pipeline target pixel file.

        Parameters
        ----------
        cube_file : str
            The cube file containing all the images to be cutout.
            Must be in the format returned by ~astrocut.make_cube.
        coordinates : str or `astropy.coordinates.SkyCoord` object
            The position around which to cutout.
            It may be specified as a string ("ra dec" in degrees)
            or as the appropriate `~astropy.coordinates.SkyCoord` object.
        cutout_size : int, array-like, `~astropy.units.Quantity`
            The size of the cutout array. If ``cutout_size``
            is a scalar number or a scalar `~astropy.units.Quantity`,
            then a square cutout of ``cutout_size`` will be created.  If
            ``cutout_size`` has two elements, they should be in ``(ny, nx)``
            order.  Scalar numbers in ``cutout_size`` are assumed to be in
            units of pixels. `~astropy.units.Quantity` objects must be in pixel or
            angular units.
        product : str
            The product type to make the cutouts from.
            Can either be 'SPOC' or 'TICA' (default is 'SPOC').
        target_pixel_file : str
            Optional. The name for the output target pixel file.
            If no name is supplied, the file will be named:
            ``<cube_file_base>_<ra>_<dec>_<cutout_size>_astrocut.fits``
        output_path : str
            Optional. The path where the output file is saved.
            The current directory is default.
        threads : int, "auto", default=1
            Number of threads to use when making remote (e.g. s3) cutouts, will not use threads for local access
            <=1 disables the threadpool, >1 sets threadpool to the specified number of threads,
            "auto" uses `concurrent.futures.ThreadPoolExecutor`'s default: cpu_count + 4, limit to max of 32
        verbose : bool
            Optional. If true intermediate information is printed.

        Returns
        -------
        response: string or None
            If successfull, returns the path to the target pixel file,
            if unsuccessful returns None.
        """

        if verbose:
            start_time = time()

        # Declare the product type being used to make the cutouts
        self.product = product.upper()

        warnings.filterwarnings("ignore", category=wcs.FITSFixedWarning)
        fits_options: Dict[str, Any] = {"mode": "denywrite", "lazy_load_hdus": True}
        if cube_file.startswith("s3://"):
            fits_options["use_fsspec"] = True
            # block size should be:
            # block_size <= m * hdul[1].section.shape[2] * hdul[1].section.shape[3] * 4 bytes
            fits_options["fsspec_kwargs"] = {"default_block_size": 10_000, "anon": True}
            # only use .section for remote data
            cube_data_prop = "section"
        else:
            fits_options["memmap"] = True
            cube_data_prop = "data"
            # disable threading for local storage access
            threads = 1
        with fits.open(cube_file, **fits_options) as cube:

            # Get the info we need from the data table
            self._parse_table_info(cube[2].data, verbose)

            if isinstance(coordinates, SkyCoord):
                self.center_coord = coordinates
            else:
                self.center_coord = SkyCoord(coordinates, unit='deg')

            if verbose:
                print("Cutout center coordinate: {},{}".format(self.center_coord.ra.deg,
                                                               self.center_coord.dec.deg))

            # Making size into an array [ny, nx]
            if np.isscalar(cutout_size):
                cutout_size = np.repeat(cutout_size, 2)

            if isinstance(cutout_size, u.Quantity):
                cutout_size = np.atleast_1d(cutout_size)
                if len(cutout_size) == 1:
                    cutout_size = np.repeat(cutout_size, 2)

            if len(cutout_size) > 2:
                warnings.warn("Too many dimensions in cutout size, only the first two will be used.",
                              InputWarning)
                cutout_size = cutout_size[:2]
                
            # Get cutout limits
            self._get_cutout_limits(cutout_size)

            if verbose:
                print("xmin,xmax: {}".format(self.cutout_lims[1]))
                print("ymin,ymax: {}".format(self.cutout_lims[0]))

            # Make the cutout
            img_cutout, uncert_cutout, aperture = self._get_cutout(getattr(cube[1], cube_data_prop), threads=threads,
                                                                   verbose=verbose)

            # Get cutout wcs info
            cutout_wcs_full = self._get_full_cutout_wcs(cube[2].header)
            max_dist, sigma = self._fit_cutout_wcs(cutout_wcs_full, img_cutout.shape[1:])
            if verbose:
                print("Maximum distance between approximate and true location: {}".format(max_dist))
                print("Error in approximate WCS (sigma): {}".format(sigma))
                
            cutout_wcs_dict = self._get_cutout_wcs_dict()
    
            # Build the TPF
            tpf_object = self._build_tpf(cube, img_cutout, uncert_cutout, cutout_wcs_dict, aperture)

            if verbose:
                write_time = time()

            if not target_pixel_file:
                _, flename = os.path.split(cube_file)

                width = self.cutout_lims[0, 1]-self.cutout_lims[0, 0]
                height = self.cutout_lims[1, 1]-self.cutout_lims[1, 0]
                target_pixel_file = "{}_{:7f}_{:7f}_{}x{}_astrocut.fits".format(flename.rstrip('.fits').rstrip("-cube"),
                                                                                self.center_coord.ra.value,
                                                                                self.center_coord.dec.value,
                                                                                width,
                                                                                height)
            target_pixel_file = os.path.join(output_path, target_pixel_file)
            
        
            if verbose:
                print("Target pixel file: {}".format(target_pixel_file))

            # Make sure the output directory exists
            if not os.path.exists(output_path):
                os.makedirs(output_path)
        
            # Write the TPF
            tpf_object.writeto(target_pixel_file, overwrite=True, checksum=True)

        if verbose:
            print("Write time: {:.2} sec".format(time()-write_time))
            print("Total time: {:.2} sec".format(time()-start_time))

        return target_pixel_file



import pytest

import numpy as np
from os import path
from re import findall

from astropy.io import fits
from astropy import wcs
from astropy.coordinates import SkyCoord
from astropy import units as u

from PIL import Image

from .utils_for_test import create_test_imgs
from .. import cutouts
from ..exceptions import InputWarning, InvalidInputError, InvalidQueryError 


@pytest.mark.parametrize('ffi_type', ['SPOC', 'TICA'])
def test_fits_cut(tmpdir, capsys, ffi_type):

    test_images = create_test_imgs(ffi_type, 50, 6, dir_name=tmpdir)

    # Single file
    center_coord = SkyCoord("150.1163213 2.200973097", unit='deg')
    cutout_size = 10
    cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    
    cutout_hdulist = fits.open(cutout_file)
    assert len(cutout_hdulist) == len(test_images) + 1  # num imgs + primary header

    cut1 = cutout_hdulist[1].data
    assert cut1.shape == (cutout_size, cutout_size)
    assert cutout_hdulist[1].data.shape == cutout_hdulist[2].data.shape
    assert cutout_hdulist[2].data.shape == cutout_hdulist[3].data.shape
    assert cutout_hdulist[3].data.shape == cutout_hdulist[4].data.shape
    assert cutout_hdulist[4].data.shape == cutout_hdulist[5].data.shape
    assert cutout_hdulist[5].data.shape == cutout_hdulist[6].data.shape
    
    cut_wcs = wcs.WCS(cutout_hdulist[1].header)
    sra, sdec = cut_wcs.all_pix2world(cutout_size/2, cutout_size/2, 0)
    assert round(float(sra), 4) == round(center_coord.ra.deg, 4)
    assert round(float(sdec), 4) == round(center_coord.dec.deg, 4)

    cutout_hdulist.close()

    # Multiple files
    cutout_files = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=False, output_dir=tmpdir)

    assert isinstance(cutout_files, list)
    assert len(cutout_files) == len(test_images)

    cutout_hdulist = fits.open(cutout_files[0])
    assert len(cutout_hdulist) == 2
    
    cut1 = cutout_hdulist[1].data
    assert cut1.shape == (cutout_size, cutout_size)
    
    cut_wcs = wcs.WCS(cutout_hdulist[1].header)
    sra, sdec = cut_wcs.all_pix2world(cutout_size/2, cutout_size/2, 0)
    assert round(float(sra), 4) == round(center_coord.ra.deg, 4)
    assert round(float(sdec), 4) == round(center_coord.dec.deg, 4)

    cutout_hdulist.close()

    # Output directory that has to be made
    cutout_files = cutouts.fits_cut(test_images, center_coord, cutout_size,
                                    output_dir=path.join(tmpdir, "cutout_files"), single_outfile=False)

    assert isinstance(cutout_files, list)
    assert len(cutout_files) == len(test_images)
    assert "cutout_files" in cutout_files[0]

    # Memory only flag
    cutout_list = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True, memory_only=True)
    assert isinstance(cutout_list, list)
    assert len(cutout_list) == 1
    assert isinstance(cutout_list[0], fits.HDUList)
    
    # Do an off the edge test
    center_coord = SkyCoord("150.1163213 2.2005731", unit='deg')
    cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    
    cutout_hdulist = fits.open(cutout_file)
    assert len(cutout_hdulist) == len(test_images) + 1  # num imgs + primary header

    cut1 = cutout_hdulist[1].data
    assert cut1.shape == (cutout_size, cutout_size)
    assert np.isnan(cut1[:cutout_size//2, :]).all()

    cutout_hdulist.close()

    # Test when the requested cutout is not on the image
    center_coord = SkyCoord("140.1163213 2.2005731", unit='deg')
    with pytest.raises(Exception, match='Cutout location is not in image footprint!') as e:
        cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True)
        assert e.type is InvalidQueryError

    center_coord = SkyCoord("15.1163213 2.2005731", unit='deg')
    with pytest.raises(Exception, match='Cutout location is not in image footprint!') as e:
        cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True)
        assert e.type is InvalidQueryError
    

    # Test when cutout is in some images not others

    # Putting zeros into 2 images
    for img in test_images[:2]:
        hdu = fits.open(img, mode="update")
        hdu[0].data[:20, :] = 0
        hdu.flush()
        hdu.close()
        
        
    center_coord = SkyCoord("150.1163213 2.2007", unit='deg')
    cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True, output_dir=tmpdir)
    
    cutout_hdulist = fits.open(cutout_file)
    assert len(cutout_hdulist) == len(test_images) + 1  # num imgs + primary header
    assert (cutout_hdulist[1].data == 0).all()
    assert (cutout_hdulist[2].data == 0).all()
    assert ~(cutout_hdulist[3].data == 0).any()
    assert ~(cutout_hdulist[4].data == 0).any()
    assert ~(cutout_hdulist[5].data == 0).any()
    assert ~(cutout_hdulist[6].data == 0).any()

    
    cutout_files = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=False, output_dir=tmpdir)
    assert isinstance(cutout_files, list)
    assert len(cutout_files) == len(test_images) - 2

    # Test when cutout is in no images
    for img in test_images[2:]:
        hdu = fits.open(img, mode="update")
        hdu[0].data[:20, :] = 0
        hdu.flush()
        hdu.close()

    with pytest.raises(Exception) as e:
        cutout_file = cutouts.fits_cut(test_images, center_coord, cutout_size, single_outfile=True, output_dir=tmpdir)
        assert e.type is InvalidQueryError
        assert "Cutout contains no data! (Check image footprint.)" in str(e.value)

    # test single image and also conflicting sip keywords
    test_image = create_test_imgs(ffi_type, 50, 1, dir_name=tmpdir,
                                  basename="img_badsip_{:04d}.fits", bad_sip_keywords=True)[0]

    center_coord = SkyCoord("150.1163213 2.2007", unit='deg')
    cutout_size = [10, 15]
    cutout_file = cutouts.fits_cut(test_image, center_coord, cutout_size, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    assert "10-x-15" in cutout_file
    cutout_hdulist = fits.open(cutout_file)
    assert cutout_hdulist[1].data.shape == (15, 10)

    center_coord = SkyCoord("150.1159 2.2006", unit='deg')
    cutout_size = [10, 15]*u.pixel
    cutout_file = cutouts.fits_cut(test_image, center_coord, cutout_size, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    assert "10.0pix-x-15.0pix" in cutout_file
    cutout_hdulist = fits.open(cutout_file)
    assert cutout_hdulist[1].data.shape == (15, 10)

    cutout_size = [1, 2]*u.arcsec
    cutout_file = cutouts.fits_cut(test_image, center_coord, cutout_size, output_dir=tmpdir, verbose=True)
    assert isinstance(cutout_file, str)
    assert "1.0arcsec-x-2.0arcsec" in cutout_file
    cutout_hdulist = fits.open(cutout_file)
    assert cutout_hdulist[1].data.shape == (33, 17)
    captured = capsys.readouterr()
    assert "Original image shape: (50, 50)" in captured.out
    assert "Image cutout shape: (33, 17)" in captured.out
    assert "Total time:" in captured.out

    center_coord = "150.1159 2.2006"
    cutout_size = [10, 15, 20]
    with pytest.warns(InputWarning):
        cutout_file = cutouts.fits_cut(test_image, center_coord, cutout_size, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    assert "10-x-15" in cutout_file
    assert "x-20" not in cutout_file

    # Test single cloud image
    test_s3_uri = "s3://stpubdata/hst/public/j8pu/j8pu0y010/j8pu0y010_drc.fits"
    center_coord = SkyCoord("150.4275416667 2.42155", unit='deg')
    cutout_size = [10, 15]
    cutout_file = cutouts.fits_cut(test_s3_uri, center_coord, cutout_size, output_dir=tmpdir)
    assert isinstance(cutout_file, str)
    assert "10-x-15" in cutout_file
    
    with fits.open(cutout_file) as cutout_hdulist:
        assert cutout_hdulist[1].data.shape == (15, 10)


def test_normalize_img():

    # basic linear stretch
    img_arr = np.array([[1, 0], [.25, .75]])
    assert ((img_arr*255).astype(int) == cutouts.normalize_img(img_arr, stretch='linear')).all()

    # invert
    assert (255-(img_arr*255).astype(int) == cutouts.normalize_img(img_arr, stretch='linear', invert=True)).all()

    # linear stretch where input image must be scaled 
    img_arr = np.array([[10, 5], [2.5, 7.5]])
    norm_img = ((img_arr - img_arr.min())/(img_arr.max()-img_arr.min())*255).astype(int)
    assert (norm_img == cutouts.normalize_img(img_arr, stretch='linear')).all()

    # min_max val
    minval, maxval = 0, 1
    img_arr = np.array([[1, 0], [-1, 2]])
    norm_img = cutouts.normalize_img(img_arr, stretch='linear', minmax_value=[minval, maxval])
    img_arr[img_arr < minval] = minval
    img_arr[img_arr > maxval] = maxval
    assert ((img_arr*255).astype(int) == norm_img).all()

    minval, maxval = 0, 1
    img_arr = np.array([[1, 0], [.1, .2]])
    norm_img = cutouts.normalize_img(img_arr, stretch='linear', minmax_value=[minval, maxval])
    img_arr[img_arr < minval] = minval
    img_arr[img_arr > maxval] = maxval
    ((img_arr*255).astype(int) == norm_img).all()

    # min_max percent
    img_arr = np.array([[1, 0], [0.1, 0.9], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr, stretch='linear', minmax_percent=[25, 75])
    assert (norm_img == [[255, 0], [0, 255], [39, 215]]).all()

    # asinh
    img_arr = np.array([[1, 0], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr)
    assert ((np.arcsinh(img_arr*10)/np.arcsinh(10)*255).astype(int) == norm_img).all()

    # sinh
    img_arr = np.array([[1, 0], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr, stretch='sinh')
    assert ((np.sinh(img_arr*3)/np.sinh(3)*255).astype(int) == norm_img).all()

    # sqrt
    img_arr = np.array([[1, 0], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr, stretch='sqrt')
    assert ((np.sqrt(img_arr)*255).astype(int) == norm_img).all()

    # log
    img_arr = np.array([[1, 0], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr, stretch='log')
    assert ((np.log(img_arr*1000+1)/np.log(1000)*255).astype(int) == norm_img).all()

    # Bad stretch
    with pytest.raises(InvalidInputError):
        img_arr = np.array([[1, 0], [.25, .75]])
        cutouts.normalize_img(img_arr, stretch='lin')

    # Giving both minmax percent and cut
    img_arr = np.array([[1, 0], [.25, .75]])
    norm_img = cutouts.normalize_img(img_arr, stretch='asinh', minmax_percent=[0.7, 99.3])
    with pytest.warns(InputWarning):
        test_img = cutouts.normalize_img(img_arr, stretch='asinh', minmax_value=[5, 2000], minmax_percent=[0.7, 99.3])
    assert (test_img == norm_img).all()


@pytest.mark.parametrize('ffi_type', ['SPOC', 'TICA'])
def test_img_cut(tmpdir, capsys, ffi_type):

    test_images = create_test_imgs(ffi_type, 50, 6, dir_name=tmpdir)
    center_coord = SkyCoord("150.1163213 2.200973097", unit='deg')
    cutout_size = 10

    # Basic jpg image
    jpg_files = cutouts.img_cut(test_images, center_coord, cutout_size, output_dir=tmpdir)
    
    assert len(jpg_files) == len(test_images)
    with open(jpg_files[0], 'rb') as IMGFLE:
        assert IMGFLE.read(3) == b'\xFF\xD8\xFF'  # JPG

    # Png (single input file, not as list)
    img_files = cutouts.img_cut(test_images[0], center_coord, cutout_size, img_format='png', output_dir=tmpdir)
    with open(img_files[0], 'rb') as IMGFLE:
        assert IMGFLE.read(8) == b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'  # PNG
    assert len(img_files) == 1

    # Color image
    color_jpg = cutouts.img_cut(test_images[:3], center_coord, cutout_size, colorize=True, output_dir=tmpdir)
    img = Image.open(color_jpg)
    assert img.mode == 'RGB'

    # Too few input images
    with pytest.raises(InvalidInputError):
        cutouts.img_cut(test_images[0], center_coord, cutout_size, colorize=True, output_dir=tmpdir)

    # Too many input images
    with pytest.warns(InputWarning):
        color_jpg = cutouts.img_cut(test_images, center_coord, cutout_size, colorize=True, output_dir=tmpdir)
    img = Image.open(color_jpg)
    assert img.mode == 'RGB'

    # string coordinates and verbose
    center_coord = "150.1163213 2.200973097"
    jpg_files = cutouts.img_cut(test_images, center_coord, cutout_size,
                                output_dir=path.join(tmpdir, "image_path"), verbose=True)
    captured = capsys.readouterr()
    assert len(findall("Original image shape", captured.out)) == 6
    assert "Cutout fits file(s)" in captured.out
    assert "Total time" in captured.out

    # test color image where one of the images is all zeros
    hdu = fits.open(test_images[0], mode='update')
    hdu[0].data[:, :] = 0
    hdu.flush()
    hdu.close()

    with pytest.raises(InvalidInputError):
        cutouts.img_cut(test_images[:3], center_coord, cutout_size,
                        colorize=True, img_format='png', output_dir=tmpdir)

    
    



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module implements the functionality to create image cubes for the purposes of
creating cutout target pixel files.
"""

import os
from copy import deepcopy
from datetime import date
from sys import platform, version_info
from time import time

import numpy as np
from astropy.io import fits
from astropy.table import Column, Table

if (version_info >= (3, 8)) and (platform != "win32"):
    from mmap import MADV_SEQUENTIAL

__all__ = ['CubeFactory', 'TicaCubeFactory']
ERROR_MSG = "One or more incorrect file types were input. Please input TICA FFI files when using\
                   ``TicaCubeFactory``, and SPOC FFI files when using ``CubeFactory``."


class CubeFactory():
    """
    Class for creating image cubes.
    
    This class emcompasses all of the cube making functionality.  
    In the current version this means creating image cubes fits files from TESS full frame image sets.
    Future versions will include more generalized cubing functionality.
    """

    def __init__(self, max_memory=50):
        """ Setting up the class members """

        self.max_memory = max_memory  # in GB
        self.block_size = None  # Number of rows
        self.num_blocks = None
        self.cube_shape = None
        
        self.time_keyword = 'TSTART'  # TESS-specific
        self.last_file_keywords = ['DATE-END', 'TSTOP']  # TESS-specific (assumed to be in extension 1)
        self.image_header_keywords = ['CAMERA', 'CCD']  # TESS-specific
        self.template_requirements = {"WCSAXES": 2}  # TESS-specific (assumed to be in extension 1)

        self.file_list = None
        self.template_file = None
        
        self.primary_header = None
        self.info_table = None
        self.cube_file = None

        
    def _configure_cube(self, file_list, **extra_keywords):
        """ 
        Run through all the files and set up the  basic parameters for the cube.
        Set up the cube primary header.
        """
        
        file_list = np.array(file_list)
        image_shape = None
        start_times = np.zeros(len(file_list))
        for i, ffi in enumerate(file_list):

            ffi_data = fits.open(ffi, mode='denywrite', memmap=True)
            
            start_times[i] = ffi_data[1].header.get(self.time_keyword)

            if image_shape is None:  # Only need to fill this once
                image_shape = ffi_data[1].data.shape
            
            if self.template_file is None:  # Only check this if we don't already have it

                is_template = True
                for key, value in self.template_requirements.items():
                    if ffi_data[1].header.get(key) != value:  # Checking for a good image header
                        is_template = False
                        
                if is_template:
                    self.template_file = ffi
                    
            ffi_data.close()
                
        self.file_list = file_list[np.argsort(start_times)]

        # Working out the block size and number of blocks needed for writing the cube
        # without using too much memory
        try:
            slice_size = image_shape[1] * len(self.file_list) * 2 * 4  # in bytes (float32)
        except IndexError:
            raise ValueError(ERROR_MSG)
        max_block_size = int((self.max_memory * 1e9)//slice_size)
        
        self.num_blocks = int(image_shape[0]/max_block_size + 1)
        self.block_size = int(image_shape[0]/self.num_blocks + 1)
        self.cube_shape = (image_shape[0], image_shape[1], len(self.file_list), 2)

        # Making the primary header
        with fits.open(self.file_list[0], mode='denywrite', memmap=True) as first_file:
            header = deepcopy(first_file[0].header)
            header.remove('CHECKSUM', ignore_missing=True)

            # Adding standard keywords
            header['ORIGIN'] = 'STScI/MAST'
            header['DATE'] = str(date.today())

            # Adding factory specific keywords
            for kwd in self.image_header_keywords:
                header[kwd] = (first_file[1].header[kwd], first_file[1].header.comments[kwd])

            # Adding the extra keywords passed in
            for kwd, value in extra_keywords.items():
                header[kwd] = (value[0], value[1])
                
        # Adding the keywords from the last file
        with fits.open(self.file_list[-1], mode='denywrite', memmap=True) as last_file:
            for kwd in self.last_file_keywords:
                header[kwd] = (last_file[1].header[kwd], last_file[1].header.comments[kwd])

        header["EXTNAME"] = "PRIMARY"

        self.primary_header = header


    def _build_info_table(self):
        """
        Reading the keywords and setting up the table to hold the image headers (extension 1)
        from every input file.
        """

        with fits.open(self.template_file, mode='denywrite', memmap=True) as ffi_data:
            
            # The image specific header information will be saved in a table in the second extension
            secondary_header = ffi_data[1].header

            # set up the image info table
            cols = []
            for kwd, val, cmt in secondary_header.cards: 
                if isinstance(val, str):
                    tpe = "S" + str(len(val))  # TODO: Maybe switch to U?
                elif isinstance(val, int):
                    tpe = np.int32
                else:
                    tpe = np.float64

                # Adding columns one by one 
                cols.append(Column(name=kwd, dtype=tpe, length=len(self.file_list), meta={"comment": cmt}))
                    
            cols.append(Column(name="FFI_FILE", dtype="S" + str(len(os.path.basename(self.template_file))),
                               length=len(self.file_list)))
            self.info_table = Table(cols)

        
    def _build_cube_file(self, cube_file):
        """
        Build the cube file on disk with primary header, cube extension header,
        and space for the cube, filled with zeros.

        Note, this will overwrite the file if it already exists.
        """
        
        # Making sure the output directory exists
        direc, _ = os.path.split(cube_file)
        if direc and not os.path.exists(direc):
            os.makedirs(direc)

        # Writing the primary header
        self.primary_header.tofile(cube_file, overwrite=True)

        # Making the cube header and writing it
        data = np.zeros((100, 100, 10, 2), dtype=np.float32)
        hdu = fits.ImageHDU(data)
        header = hdu.header
        header["NAXIS4"], header["NAXIS3"], header["NAXIS2"], header["NAXIS1"] = self.cube_shape

        with open(cube_file, 'ab') as CUBE:
            CUBE.write(bytearray(header.tostring(), encoding="utf-8"))

        # Expanding the file to fit the full data cube
        # fits requires all blocks to be a multiple of 2880
        cubesize_in_bytes = ((np.prod(self.cube_shape) * 4 + 2880 - 1) // 2880) * 2880
        filelen = os.path.getsize(cube_file)
        with open(cube_file, 'r+b') as CUBE:
            CUBE.seek(filelen + cubesize_in_bytes - 1)
            CUBE.write(b'\0')

        self.cube_file = cube_file


    def _write_block(self, cube_hdu, start_row=0, end_row=None, fill_info_table=False, verbose=False):
        """
        Do one pass through the image files and write a block of the cube.

        cube_hdu is am hdulist object opened in update mode
        """

        # Initializing the sub-cube
        nrows = (self.cube_shape[0] - start_row) if (end_row is None) else (end_row - start_row)
        sub_cube = np.zeros((nrows, *self.cube_shape[1:]), dtype=np.float32)
        
        # Loop through files
        for i, fle in enumerate(self.file_list):

            if verbose:
                st = time()

            with fits.open(fle, mode='denywrite', memmap=True) as ffi_data:

                # add the image and info to the arrays
                sub_cube[:, :, i, 0] = ffi_data[1].data[start_row:end_row, :]
                sub_cube[:, :, i, 1] = ffi_data[2].data[start_row:end_row, :]

                del ffi_data[1].data
                del ffi_data[2].data
                
                if fill_info_table:  # Also save the header info in the info table

                    for kwd in self.info_table.columns:
                        if kwd == "FFI_FILE":
                            self.info_table[kwd][i] = os.path.basename(fle)
                        else:
                            nulval = None
                            if self.info_table[kwd].dtype.name == "int32":
                                nulval = 0
                            elif self.info_table[kwd].dtype.char == "S":  # hacky way to check if it's a string
                                nulval = ""
                            self.info_table[kwd][i] = ffi_data[1].header.get(kwd, nulval)

            if verbose:
                print(f"Completed file {i} in {time()-st:.3} sec.")

        # Fill block and flush to disk
        cube_hdu[1].data[start_row:end_row, :, :, :] = sub_cube

        if (version_info <= (3, 8)) or (platform == "win32"):
            cube_hdu.flush()

        del sub_cube

        
    def _write_info_table(self):
        """
        Turn the info table into an HDU object and append it to the cube file
        """

        # Make table hdu 
        cols = []
        for kwd in self.info_table.columns:
            if self.info_table[kwd].dtype == np.float64:
                tpe = 'D'
            elif self.info_table[kwd].dtype == np.int32:
                tpe = 'J'
            else:
                tpe = str(self.info_table[kwd].dtype).replace("S", "A").strip("|")
        
            cols.append(fits.Column(name=kwd, format=tpe, array=self.info_table[kwd]))
        
        col_def = fits.ColDefs(cols)
        table_hdu = fits.BinTableHDU.from_columns(col_def)

        # Adding the comments to the header
        for kwd in self.info_table.columns:
            if kwd in ['XTENSION', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'PCOUNT', 'GCOUNT', 'TFIELDS']:
                continue  # skipping the keyword already in use
            table_hdu.header[kwd] = self.info_table[kwd].meta.get("comment", "")

        # Appending to the cube file
        with fits.open(self.cube_file, mode='append', memmap=True) as cube_hdus:
            cube_hdus.append(table_hdu)


    
    def make_cube(self, file_list, cube_file="img-cube.fits", sector=None, max_memory=50, verbose=True):
        """
        Turns a list of fits image files into one large data-cube.
        Input images must all have the same footprint and resolution.
        The resulting datacube is transposed for quicker cutouts.
        This function can take some time to run, exactly how much time will depend on the number
        of input files and the maximum allowed memory. The runtime will be fastest if the
        entire data cube can be held in memory, however that can be quite large (~40GB for a full
        TESS main mission sector, 3 times that for a TESS extended mission sector).

        Parameters
        ----------
        file_list : array
            The list of fits image files to cube.  
            Assumed to have the format of a TESS FFI:
            - A primary HDU consisting only of a primary header
            - An image HDU containing the image
            - A second image HDU containing the uncertainty image
        cube_file : string
            Optional.  The filename/path to save the output cube in. 
        sector : int
            Optional.  TESS sector to add as header keyword (not present in FFI files).
        max_memory : float
            Optional, default is 50. The maximum amount of memory to make available for building
            the data cube in GB. Note, this is the maximum amount of space to be used for the cube
            array only, so should not be set to the full amount of memory on the system.
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response: string or None
            If successful, returns the path to the cube fits file, 
            if unsuccessful returns None.
        """

        if verbose:
            startTime = time()

        self.max_memory = max_memory

        # Set up the basic cube parameters
        sector = (sector, "Observing sector")

        self._configure_cube(file_list, sector=sector)
    
        if verbose:
            print("Using {} to initialize the image header table.".format(os.path.basename(self.template_file)))
            print(f"Cube will be made in {self.num_blocks} blocks of {self.block_size} rows each.")

        # Set up the table to old the individual image heades
        self._build_info_table()

        
        # Write the empty file, ready for the cube to be added
        self._build_cube_file(cube_file)

        # Fill the image cube
        with fits.open(self.cube_file, mode='update', memmap=True) as cube_hdu:

            if (version_info >= (3, 8)) and (platform != "win32"):
                mm = fits.util._get_array_mmap(cube_hdu[1].data)
                mm.madvise(MADV_SEQUENTIAL)

            for i in range(self.num_blocks):
                start_row = i * self.block_size
                end_row = start_row + self.block_size
                if end_row >= self.cube_shape[0]:
                    end_row = None

                fill_info_table = True if (i == 0) else False
                self._write_block(cube_hdu, start_row, end_row, fill_info_table, verbose)
                  
                if verbose:
                    print(f"Completed block {i+1} of {self.num_blocks}")

        # Add the info table to the cube file
        self._write_info_table()
        if verbose:
            print(f"Total time elapsed: {(time() - startTime)/60:.2f} min")

        return self.cube_file


class TicaCubeFactory():
    """
    Class for creating TICA image cubes.
    
    This class emcompasses all of the cube making functionality.  
    In the current version this means creating image cubes fits files from TESS full frame image sets.
    Future versions will include more generalized cubing functionality.

    The TESS Image CAlibrator (TICA) products are high level science products (HLSPs) 
    developed by the MIT Quick Look Pipeline (https://github.com/mmfausnaugh/tica). These 
    images are produced and delivered up to 4x sooner than their SPOC counterparts (as of TESS EM2),
    and can therefore be used to produce the most up-to-date cutouts of a target. 
    More information on TICA can be found here: https://archive.stsci.edu/hlsp/tica
    """

    def __init__(self, max_memory=50):
        """ Setting up the class members."""

        self.max_memory = max_memory  # in GB
        self.block_size = None  # Number of rows
        self.num_blocks = None
        self.cube_shape = None
        
        self.time_keyword = 'STARTTJD'  # Start time in TJD. TICA-specific.
        self.last_file_keywords = ['ENDTJD']  # Stop time in TJD. TICA-specific (assumed to be in extension 0)                  
        self.image_header_keywords = ['CAMNUM', 'CCDNUM']  # Camera number and CCD number 
        self.template_requirements = {'NAXIS': 2}  # Using NAXIS instead of WCSAXES. 
                                                   
        self.file_list = None
        self.template_file = None
        
        self.primary_header = None
        self.info_table = None
        self.cube_file = None

        # Used when updating an existing cube 
        self.old_cols = None
        self.update = False
        self.cube_append = None

    def _configure_cube(self, file_list, **extra_keywords):
        """ Run through all the files and set up the basic parameters for the cube.
        Set up the cube primary header.
        """
        
        file_list = np.array(file_list)
        image_shape = None
        start_times = np.zeros(len(file_list))
        for i, ffi in enumerate(file_list):
            ffi_data = fits.open(ffi, mode='denywrite', memmap=True)
            
            start_times[i] = ffi_data[0].header.get(self.time_keyword)

            if image_shape is None:  # Only need to fill this once
                try:
                    image_shape = ffi_data[0].data.shape
                except AttributeError:
                    raise ValueError(ERROR_MSG)
            
            if self.template_file is None:  # Only check this if we don't already have it

                is_template = True
                for key, value in self.template_requirements.items():
                    if ffi_data[0].header.get(key) != value:  # Checking for a good image header
                        is_template = False
                        
                if is_template:
                    self.template_file = ffi
                    
            ffi_data.close()

        self.file_list = file_list[np.argsort(start_times)]

        # Working out the block size and number of blocks needed for writing the cube
        # without using too much memory
        slice_size = image_shape[1] * len(self.file_list) * 2 * 4  # in bytes (float32)
        max_block_size = (self.max_memory * 1e9) // slice_size
        
        self.num_blocks = int(image_shape[0]/max_block_size + 1)
        self.block_size = int(image_shape[0]/self.num_blocks + 1)
        
        # Determining cube shape: 
        # If it's a new TICA cube, the shape is (nRows, nCols, nImages, 1).
        # Axis 4 is `1` instead of `2` because we do not work with error arrays for TICA.
        if not self.update:
            self.cube_shape = (image_shape[0], image_shape[1], len(self.file_list), 1)
            
        # Else, if it's an update to an existing cube, the shape is (nRows, nCols, nImages + nNewImages, 2)
        else: 
            self.cube_shape = self.cube_append.shape
            
        # Making the primary header if there's no cube_file yet
        if not self.update:
            with fits.open(self.file_list[0], mode='denywrite', memmap=True) as first_file:
                header = deepcopy(first_file[0].header)
                header.remove('CHECKSUM', ignore_missing=True)

                # Adding standard keywords
                header['ORIGIN'] = 'STScI/MAST'
                header['DATE'] = str(date.today())

                # Adding factory specific keywords
                for kwd in self.image_header_keywords:
                    # TICA file structure differs from SPOC in that factory-specific kwds
                    # are in the 0th extension, along with the science data.
                    header[kwd] = (first_file[0].header[kwd], first_file[0].header.comments[kwd])

                # Adding the extra keywords passed in
                for kwd, value in extra_keywords.items():
                    header[kwd] = (value[0], value[1])

        # Otherwise we're updating an existing cube file
        else:
            with fits.open(self.cube_file, mode='update', memmap=True) as cube_hdu:
                header = cube_hdu[0].header 
                header['HISTORY'] = f'Updated on {str(date.today())} with new FFI delivery.' 
                header['HISTORY'] = f'First FFI is {str(os.path.basename(self.file_list[0]))}'

        # Adding the keywords from the last file
        with fits.open(self.file_list[-1], mode='denywrite', memmap=True) as last_file:
            for kwd in self.last_file_keywords:
                # TICA file structure differs from SPOC in that factory-specific kwds
                # are in the 0th extension, along with the science data.
                header[kwd] = (last_file[0].header[kwd], last_file[0].header.comments[kwd])

        header["EXTNAME"] = "PRIMARY"

        self.primary_header = header

    def _build_info_table(self):
        """ Reading the keywords and setting up the table to hold the image headers (extension 1)
        from every input file.
        """

        with fits.open(self.template_file, mode='denywrite', memmap=True) as ffi_data:
            
            # The image specific header information will be saved in a table in the second extension
            primary_header = ffi_data[0].header

            # set up the image info table
            existing_cols = []
            cols = []
            length = len(self.file_list)
            for kwd, val, cmt in primary_header.cards: 
                if isinstance(val, str):  
                    tpe = "S" + str(len(val))  # TODO: Maybe switch to U?
                elif isinstance(val, int):
                    tpe = np.int32
                else:
                    tpe = np.float64
                    
                if kwd not in existing_cols:
                    existing_cols.append(kwd)
                    cols.append(Column(name=kwd, dtype=tpe, length=length, meta={"comment": cmt}))

            cols.append(Column(name="FFI_FILE", dtype="S" + str(len(os.path.basename(self.template_file))),
                               length=length))
            
            self.info_table = Table(cols)

    def _build_cube_file(self, cube_file):
        """ Build the cube file on disk with primary header, cube extension header,
        and space for the cube, filled with zeros.

        Note, this will overwrite the file if it already exists.

        Parameters
        ----------
        cube_file : str 
            Optional. The filename/path to save the output cube in. 
        """
        
        # Making sure the output directory exists
        direc, _ = os.path.split(cube_file)
        if direc and not os.path.exists(direc):
            os.makedirs(direc)

        # Writing the primary header
        hdu0 = fits.PrimaryHDU(header=self.primary_header)
        hdul = fits.HDUList([hdu0])
        hdul.writeto(cube_file, overwrite=True)

        # Making the cube header and writing it
        data = np.zeros((100, 100, 10, 2), dtype=np.float32)
        hdu = fits.ImageHDU(data)
        header = hdu.header
        header["NAXIS4"], header["NAXIS3"], header["NAXIS2"], header["NAXIS1"] = self.cube_shape

        # Writes the header into the cube as an array of bytes
        with open(cube_file, 'ab') as CUBE:
            CUBE.write(bytearray(header.tostring(), encoding="utf-8"))

        # Expanding the file to fit the full data cube
        # fits requires all blocks to be a multiple of 2880
        cubesize_in_bytes = ((np.prod(self.cube_shape) * 4 + 2880 - 1) // 2880) * 2880
        filelen = os.path.getsize(cube_file)
        
        # Seek to end of file and write null byte
        with open(cube_file, 'r+b') as CUBE:
            CUBE.seek(filelen + cubesize_in_bytes - 1)
            CUBE.write(b'\0')
        

        self.cube_file = cube_file

    def _write_block(self, cube_hdu, start_row=0, end_row=None, fill_info_table=False, verbose=False):
        """ Do one pass through the image files and write a block of the cube.

        cube_hdu is an hdulist object opened in update mode
        """

        # Initializing the sub-cube
        nrows = (self.cube_shape[0] - start_row) if (end_row is None) else (end_row - start_row)
        sub_cube = np.zeros((nrows, *self.cube_shape[1:]), dtype=np.float32)
        
        # Loop through files
        for i, fle in enumerate(self.file_list):

            if verbose:
                st = time()

            # In this section we will take the SCI data from self.file_list above
            # and "paste" a cutout of the full SCI array into a 4d array called 
            # sub_cube. We iterate this process until the full SCI array for each 
            # FFI is copied onto the cube. Usually the sub_cube is ~1/3 the size of 
            # the full SCI array, so there are 3 iterations. 
            # For TICA, the SCI data exists in the 0th extension. 
            with fits.open(fle, mode='denywrite', memmap=True) as ffi_data:

                # Add the image and info to the arrays
                sub_cube[:, :, i, 0] = ffi_data[0].data[start_row:end_row, :]

                del ffi_data[0].data
               
                # Also save the header info in the info table
                if fill_info_table:

                    # Iterate over every keyword in the TICA FFI primary header
                    for kwd in self.info_table.columns:
                        if kwd == "FFI_FILE":
                            self.info_table[kwd][i] = os.path.basename(fle)
                            
                        else:
                            nulval = None
                            if self.info_table[kwd].dtype.name == "int32":
                                nulval = 0
                            elif self.info_table[kwd].dtype.char == "S":  # hacky way to check if it's a string
                                nulval = ""
                            
                            # NOTE: 
                            # A header keyword ('COMMENT') in TICA returns a keyword 
                            # value in the form of a _HeaderCommentaryCard instead of a STRING. 
                            # This breaks the info table because the info table 
                            # doesn't understand what a FITS header commentary card is. 
                            # Adding a try/except is a way to catch when these instances happen 
                            # and turn the keyword value from a _HeaderCommentaryCard to a
                            # string, which is what it's meant to be in the info table. 
                            #
                            # TO-DO: Find a more elegant way to handle these stray 
                            # _HeaderCommentaryCards. 
                            
                            try:
                                self.info_table[kwd][i] = ffi_data[0].header.get(kwd, nulval)
                            except ValueError:
                                kwd_val = ffi_data[0].header.get(kwd)
                                if isinstance(kwd_val, fits.header._HeaderCommentaryCards):
                                    self.info_table[kwd][i] = str(kwd_val)
                                else:
                                    raise

            if verbose:
                print(f"Completed file {i} in {time()-st:.3} sec.")

        # Fill block and flush to disk
        if not self.update:
            cube_hdu[1].data[start_row:end_row, :, :, :] = sub_cube
        else:
            self.cube_append[start_row:end_row, :, :, :] = sub_cube

        if (version_info <= (3, 8)) or (platform == "win32"):
            cube_hdu.flush()

        del sub_cube
    
    def _update_info_table(self):
        """ Updating an existing info table with newly created
        """

        with fits.open(self.template_file, mode='denywrite', memmap=True) as ffi_data:
            
            # The image specific header information will be saved in a table in the second extension
            primary_header = ffi_data[0].header

            # set up the image info table
            cols = []
            for kwd, val, cmt in primary_header.cards: 
                if isinstance(val, str):
                    tpe = "S" + str(len(val))  # TODO: Maybe switch to U?
                elif isinstance(val, int):
                    tpe = np.int32
                else:
                    tpe = np.float64
                    
                # If there's already an info table, this means we are
                # updating a cube instead of making a new one, so expanding 
                # the length by the new FFIs (hence +self.file_list)

                with fits.open(self.cube_file, mode='readonly') as hdul:

                    og_table = hdul[2].data
                    appended_column = np.concatenate((og_table[kwd], self.info_table[kwd]))

                length = len(og_table)+len(self.info_table[kwd])  
                
                cols.append(Column(appended_column, name=kwd, dtype=tpe, length=length, meta={"comment": cmt}))

            with fits.open(self.cube_file, mode='readonly') as hdul:

                og_table = hdul[2].data
                appended_column = np.concatenate((og_table['FFI_FILE'], self.info_table['FFI_FILE']))

                str_length = str(len(os.path.basename(self.template_file)))
                cols.append(Column(appended_column, name="FFI_FILE", dtype="S" + str_length, length=length))
    
            self.info_table = Table(cols)


    def _write_info_table(self):
        """
        Turn the info table into an HDU object and append it to the cube file
        """

        # Make table hdu 
        cols = []
        for kwd in self.info_table.columns:

            if self.info_table[kwd].dtype == np.float64:
                tpe = 'D'
            elif self.info_table[kwd].dtype == np.int32:
                tpe = 'J'
            else:
                tpe = str(self.info_table[kwd].dtype).replace("S", "A").strip("|")
            
            cols.append(fits.Column(name=kwd, format=tpe, array=self.info_table[kwd]))
        
        col_def = fits.ColDefs(cols)
        table_hdu = fits.BinTableHDU.from_columns(col_def)

        # Adding the comments to the header
        for kwd in self.info_table.columns:
            if kwd in ['XTENSION', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'PCOUNT', 'GCOUNT', 'TFIELDS']:
                continue  # skipping the keyword already in use
            table_hdu.header[kwd] = self.info_table[kwd].meta.get("comment", "")

        # Appending to the cube file
        with fits.open(self.cube_file, mode='update', memmap=True) as cube_hdus:
            # If we're updating the cube, get rid of the existing table 
            # so we can replace it with the new one. 
            if self.update:
                cube_hdus.pop(index=2)
            cube_hdus.append(table_hdu)


    def _update_cube(self, file_list, cube_file, sector=None, max_memory=50, verbose=True):
        """ Updates an existing cube file if one has already been made and a new delivery is being appended to it. 
        Same functionality as make_cube(...), but working on an already existing file rather than building a new one. 
        This function will: 

        1. Create a new cube consisting of the new FFIs that will be appended to the existing cube
        2. Update primary header keywords to reflect the update to the file
        3. Expand the file size of the FITS file containing the cube, to accomodate for the updated one 
        4. Rename the file accordingly(?)

        """

        self.update = True  # we're updating!

        self.max_memory = max_memory

        if verbose:
            startTime = time()

        # Next locate the existing cube file and assign it a variable
        err_msg = 'Location of the cube file was unsuccessful. Please ensure the correct path was provided.'
        assert os.path.exists(cube_file), err_msg
        self.cube_file = cube_file

        if verbose:
            print(f'Updating cube file: {cube_file}')

        # Ensure that none of the files in file_list are in the cube already, to avoid duplicates
        in_cube = list(fits.getdata(self.cube_file, 2)['FFI_FILE'])

        # TO-DO: Add warning message instead of this verbose print stmnt.
        filtered_file_list = []
        for idx, file in enumerate(file_list): 

            if os.path.basename(file) in in_cube:
                print('File removed from list:')
                print(os.path.basename(file))

            if os.path.basename(file) not in in_cube:
                filtered_file_list.append(file)

        noffis_err_msg = 'No new FFIs found for the given sector.'
        assert len(filtered_file_list) > 0, noffis_err_msg

        if verbose: 
            print(f'{len(filtered_file_list)} new FFIs found!')
        
        # Creating an empty cube that will be appended to the existing cube
        og_cube = fits.getdata(cube_file, 1)
        dimensions = list(og_cube.shape)
        dimensions[2] = len(filtered_file_list)
        self.cube_append = np.zeros(dimensions)

        # Set up the basic cube parameters
        sector = (sector, "Observing sector")
        self._configure_cube(filtered_file_list, sector=sector)

        if verbose:
            print(f"FFIs will be appended in {self.num_blocks} blocks of {self.block_size} rows each.")
        
        # Starting a new info table from scratch with new rows
        self._build_info_table()

        # Update the image cube 
        with fits.open(self.cube_file, mode='update', memmap=True) as cube_hdu:

            # mmap: "allows you to take advantage of lower-level operating system 
            # functionality to read files as if they were one large string or array. 
            # This can provide significant performance improvements in code that 
            # requires a lot of file I/O."
            if (version_info >= (3, 8)) and (platform != "win32"):
                mm = fits.util._get_array_mmap(cube_hdu[1].data)
                # madvise: "Send advice option to the kernel about the memory region 
                # beginning at start and extending length bytes.""
                mm.madvise(MADV_SEQUENTIAL)

            for i in range(self.num_blocks):
                start_row = i * self.block_size
                end_row = start_row + self.block_size

                if end_row >= self.cube_shape[0]:
                    end_row = None

                # filling in the cube file with the new FFIs
                # the info table also gets updated here 
                fill_info_table = True
                self._write_block(cube_hdu, start_row, end_row, fill_info_table, verbose)

                if verbose:
                    print(f"Completed block {i+1} of {self.num_blocks}")
       
        # Append the new cube to the existing cube
        new_cube = np.concatenate((og_cube, self.cube_append), axis=2)

        # Add it to the HDU list 
        with fits.open(self.cube_file, mode='update') as hdul:
            
            if verbose:
                print(f'Original cube of size: {str(og_cube.shape)}')
                print(f'will now be replaced with cube of size: {str(new_cube.shape)}')
                print(f'for file ``{cube_file}``')
            hdul[1].data = new_cube

        # Appending new info table to original 
        self._update_info_table()
        
        # Writing the info table to EXT2 of the FITS file 
        self._write_info_table()

        if verbose:
            print(f"Total time elapsed: {(time() - startTime)/60:.2f} min")

        return self.cube_file


    def make_cube(self, file_list, cube_file="img-cube.fits", sector=None, max_memory=50, verbose=True):
        """
        Analogous to CubeFactory.make_cube(...). 
        Turns a list of fits image files into one large data-cube.
        Input images must all have the same footprint and resolution.
        The resulting datacube is transposed for quicker cutouts.
        This function can take some time to run, exactly how much time will depend on the number
        of input files and the maximum allowed memory. The runtime will be fastest if the
        entire data cube can be held in memory, however that can be quite large (~40GB for a full
        TESS main mission sector, 3 times that for a TESS extended mission sector).

        Parameters
        ----------
        file_list : array
            The list of fits image files to cube.  
            Assumed to have the format of a TESS FFI:
            - A primary HDU consisting only of a primary header
            - An image HDU containing the image
            - A second image HDU containing the uncertainty image
        cube_file : string
            Optional.  The filename/path to save the output cube in. 
        sector : int
            Optional.  TESS sector to add as header keyword (not present in FFI files).
        max_memory : float
            Optional, default is 50. The maximum amount of memory to make available for building
            the data cube in GB. Note, this is the maximum amount of space to be used for the cube
            array only, so should not be set to the full amount of memory on the system.
        verbose : bool
            Optional. If true intermediate information is printed. 

        Returns
        -------
        response: string or None
            If successful, returns the path to the cube fits file, 
            if unsuccessful returns None.
        """

        if verbose:
            startTime = time()

        self.max_memory = max_memory

        # Set up the basic cube parameters
        sector = (sector, "Observing sector")
        self._configure_cube(file_list, sector=sector)

        if verbose:
            print("Using {} to initialize the image header table.".format(os.path.basename(self.template_file)))
            print(f"Cube will be made in {self.num_blocks} blocks of {self.block_size} rows each.")
        
        # Set up the table to hold the individual image headers
        self._build_info_table()
        
        # Write the empty file, ready for the cube to be added
        self._build_cube_file(cube_file)
        
        # Fill the image cube
        with fits.open(self.cube_file, mode='update', memmap=True) as cube_hdu:
            
            if (version_info >= (3, 8)) and (platform != "win32"):
                mm = fits.util._get_array_mmap(cube_hdu[1].data)
                mm.madvise(MADV_SEQUENTIAL)

            for i in range(self.num_blocks):
                start_row = i * self.block_size
                end_row = start_row + self.block_size

                if end_row >= self.cube_shape[0]:
                    end_row = None

                fill_info_table = True if (i == 0) else False
                self._write_block(cube_hdu, start_row, end_row, fill_info_table, verbose)
                  
                if verbose:
                    print(f"Completed block {i+1} of {self.num_blocks}")

        # Add the info table to the cube file
        self._write_info_table()
        if verbose:
            print(f"Total time elapsed: {(time() - startTime)/60:.2f} min")

        return self.cube_file
        




import numpy as np
import os
import pytest

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astroquery.mast import Observations
from re import findall
from os import path

from .utils_for_test import create_test_ffis
from ..make_cube import CubeFactory, TicaCubeFactory


def test_make_cube(tmpdir):
    """
    Testing the make cube functionality by making a bunch of test FFIs, 
    making the cube, and checking the results.
    """

    cube_maker = CubeFactory()
    
    img_sz = 10
    num_im = 100
    
    ffi_files = create_test_ffis(img_sz, num_im, dir_name=tmpdir)
    cube_file = cube_maker.make_cube(ffi_files, path.join(tmpdir, "out_dir", "test_cube.fits"), verbose=False)
    hdu = fits.open(cube_file)
    cube = hdu[1].data
    
    # expected values for cube
    ecube = np.zeros((img_sz, img_sz, num_im, 2))
    plane = np.arange(img_sz*img_sz, dtype=np.float32).reshape((img_sz, img_sz))
    assert cube.shape == ecube.shape, "Mismatch between cube shape and expected shape"

    for i in range(num_im):
        ecube[:, :, i, 0] = -plane
        ecube[:, :, i, 1] = plane
        plane += img_sz*img_sz
    assert np.all(cube == ecube), "Cube values do not match expected values"
    
    tab = Table(hdu[2].data)
    assert np.all(tab['TSTART'] == np.arange(num_im)), "TSTART mismatch in table"
    assert np.all(tab['TSTOP'] == np.arange(num_im)+1), "TSTOP mismatch in table"

    filenames = np.array([path.split(x)[1] for x in ffi_files])
    assert np.all(tab['FFI_FILE'] == np.array(filenames)), "FFI_FILE mismatch in table"

    hdu.close()
    

def test_make_and_update_cube(tmpdir):
    """
    Testing the make cube and update cube functionality for TICACubeFactory by making a bunch of test FFIs, 
    making the cube with first half of the FFIs, updating the same cube with the second half,
    and checking the results.
    """

    cube_maker = TicaCubeFactory()

    img_sz = 10
    num_im = 100
    
    ffi_files = create_test_ffis(img_sz, num_im, product='TICA', dir_name=tmpdir)

    cube_file = path.join(os.getcwd(), "out_dir", "test_update_cube.fits")

    # Testing make_cube
    
    cube_maker.make_cube(ffi_files[0:num_im // 2], cube_file, verbose=False)

    hdu = fits.open(cube_file)
    cube = hdu[1].data

    # expected values for cube before update_cube
    ecube = np.zeros((img_sz, img_sz, num_im // 2, 1))
    plane = np.arange(img_sz*img_sz, dtype=np.float32).reshape((img_sz, img_sz))

    assert cube.shape == ecube.shape, "Mismatch between cube shape and expected shape"

    for i in range(num_im // 2):
        ecube[:, :, i, 0] = -plane
        # we don't need to test error array because TICA doesnt come with error arrays
        # so index 1 will always be blank
        # ecube[:, :, i, 1] = plane
        plane += img_sz*img_sz

    assert np.all(cube == ecube), "Cube values do not match expected values"

    hdu.close()

    # Testing update_cube

    cube_file = cube_maker._update_cube(ffi_files[num_im // 2:], cube_file, verbose=False)

    hdu = fits.open(cube_file)
    cube = hdu[1].data
    
    # expected values for cube after update_cube
    ecube = np.zeros((img_sz, img_sz, num_im, 1))
    plane = np.arange(img_sz*img_sz, dtype=np.float32).reshape((img_sz, img_sz))

    assert cube.shape == ecube.shape, "Mismatch between cube shape and expected shape"

    for i in range(num_im):
        ecube[:, :, i, 0] = -plane
        # we don't need to test error array because TICA doesnt come with error arrays
        # so index 1 will always be blank
        # ecube[:, :, i, 1] = plane
        plane += img_sz*img_sz

    assert np.all(cube == ecube), "Cube values do not match expected values"

    tab = Table(hdu[2].data)
    assert np.all(tab['STARTTJD'] == np.arange(num_im)), "STARTTJD mismatch in table"
    assert np.all(tab['ENDTJD'] == np.arange(num_im)+1), "ENDTJD mismatch in table"

    filenames = np.array([path.split(x)[1] for x in ffi_files])
    assert np.all(tab['FFI_FILE'] == np.array(filenames)), "FFI_FILE mismatch in table"

    hdu.close()

    os.remove(cube_file)
    

def test_iteration(tmpdir, capsys):
    """
    Testing cubes made with different numbers of iterations against each other.
    """

    cube_maker = CubeFactory()

    img_sz = 1000
    num_im = 10

    ffi_files = create_test_ffis(img_sz, num_im, dir_name=tmpdir)

    # Should produce cube in single iteration
    cube_file_1 = cube_maker.make_cube(ffi_files, path.join(tmpdir, "iterated_cube_1.fits"),
                                       max_memory=0.5, verbose=True)
    captured = capsys.readouterr()
    assert len(findall("Completed block", captured.out)) == 1, "Incorrect number of iterations"
    assert len(findall("Completed file", captured.out)) == num_im, "Incorrect number of complete files"

    cube_file_2 = cube_maker.make_cube(ffi_files, path.join(tmpdir, "iterated_cube_2.fits"),
                                       max_memory=0.05, verbose=True)
    captured = capsys.readouterr()
    assert len(findall("Completed block", captured.out)) == 2, "Incorrect number of iterations"
    assert len(findall("Completed file", captured.out)) == num_im*2, "Incorrect number of complete files"
    
    hdu_1 = fits.open(cube_file_1)
    cube_1 = hdu_1[1].data

    hdu_2 = fits.open(cube_file_2)
    cube_2 = hdu_2[1].data

    assert cube_1.shape == cube_2.shape, "Mismatch between cube shape for 1 vs 2 iterations"
    assert np.all(cube_1 == cube_2), "Cubes made in 1 vs 2 iterations do not match"

    # expected values for cube
    ecube = np.zeros((img_sz, img_sz, num_im, 2))
    plane = np.arange(img_sz*img_sz, dtype=np.float32).reshape((img_sz, img_sz))
    assert cube_1.shape == ecube.shape, "Mismatch between cube shape and expected shape"

    for i in range(num_im):
        ecube[:, :, i, 0] = -plane
        ecube[:, :, i, 1] = plane
        plane += img_sz*img_sz

    assert np.all(cube_1 == ecube), "Cube values do not match expected values"


@pytest.mark.parametrize("ffi_type", ["TICA", "SPOC"])
def test_invalid_inputs(tmpdir, ffi_type):

    coordinates = SkyCoord(289.0979, -29.3370, unit="deg")

    # Assigning some variables
    target_name = "TICA FFI" if ffi_type == "TICA" else "TESS FFI"
    value_error = "One or more incorrect file types were input. Please input TICA FFI files when using\
                   ``TicaCubeFactory``, and SPOC FFI files when using ``CubeFactory``."

    # Getting TESS sector 27 observations for the given coordinate
    observations = Observations.query_criteria(coordinates=coordinates,
                                               target_name=target_name,
                                               dataproduct_type="image",
                                               sequence_number=27)
    
    # Getting a list of products. Keeping it small so we don't have to download so many.
    products = Observations.get_product_list(observations[0])[:2]

    manifest = Observations.download_products(products, download_dir=str(tmpdir))

    if ffi_type == "TICA":
        cube_maker = CubeFactory()
    elif ffi_type == "SPOC":
        cube_maker = TicaCubeFactory()

    with pytest.raises(ValueError) as error_msg:
        cube_maker.make_cube(manifest["Local Path"])
    assert value_error in str(error_msg.value)

    

    

    
